//
// Copyright (c) 1999, Bernhard Tschirren
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of the source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//
// 2. Any redistribution solely in binary form must conspicuously
//    reproduce the following disclaimer in documentation provided with
//    the binary redistribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'', WITHOUT ANY WARRANTIES, EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  LICENSOR SHALL
// NOT BE LIABLE FOR ANY LOSS OR DAMAGES RESULTING FROM THE USE OF THIS
// SOFTWARE, EITHER ALONE OR IN COMBINATION WITH ANY OTHER SOFTWARE.
//

#include "application/Intersection.h"
#include "nodes/CellNode.h"
#include "nodes/PortalNode.h"
#include "util/Debug.h"


static Boolean lineIntersectsPortal(const Point& prevPos, const Point& curPos, const PortalNode& portal)
  {
  const Plane& plane = portal.getPlane();
  Float prevDot      = prevPos.dotProduct(plane.getNormal());
  Float curDot       = curPos.dotProduct(plane.getNormal());

  // Previous position must be infront of plane
  if (prevDot >= plane.getDistance())
    return FALSE;

  // Current position must be behind plane
  if (curDot < plane.getDistance())
    return FALSE;

  // OK, now find the point of intersection with the plane
  Float time         = (plane.getDistance() - prevDot) / (curDot - prevDot);
  Point intersection = Point(prevPos).interpolate(curPos, time);

  // Finally, verify that the point lies inside the portal polygon
  const TArray<Point>& vertices    = portal.getVertices();
  const Int            numVertices = vertices.getLength();
  for (Int i = 0; i < numVertices; i++)
    {
    // Calculate vector from v[1] to v[i+1]
    Vector edge1 = vertices[(i+1) % numVertices] - vertices[i];

    // Calculate vector from intersection to v[i+1]
    Vector edge2 = vertices[(i+1) % numVertices] - intersection;

    // Angle between these two edges MUST be < 180. Otherwise
    // point is outside the convex polygon.
    if (edge1.dotProduct(edge2) < 0)
      return FALSE;
    }

  return TRUE;
  }


Void checkPortalIntersections(RenderNode* node, const Point& prevPos)
  {
  CellNode*   srcCell;
  CellNode*   dstCell;
  PortalNode* portal;

  // Check to make sure we even have a node!
  if (node == NULL)
	return;

  // Get the first parent of the node
  srcCell = PTR_CAST(CellNode, node->getParent());
  if (srcCell == NULL)
    return;

  // Test for intersection with all portals  
  const GraphNodeArray& children = srcCell->getChildren();
  for (Int i = 0; i < children.getLength(); i++)
    {
    // Do we have a portal?
    portal = PTR_CAST(PortalNode, children[i]);
    if (portal != NULL)
      {
      // Get cell on other side of portal
      dstCell = PTR_CAST(CellNode, portal->getChild());
      if (dstCell == NULL)
        return;

      if (lineIntersectsPortal(prevPos, node->getTranslation(), *portal))
        {
        // OK, the node intersected with the portal. We need to
        // disconnect the node from the current cell, and attach
        // it to the cell on the other side of the portal.

        // Disconnect
        srcCell->detachChild(node);

        // Reconnect
        dstCell->attachChild(node);

        // Assume that we can only step through one portal at a time.
        // We do this becase attaching/detaching messes around
        // with the childen array, which we are currently looping
        // through!
        return;
        }
      }
    }
  }