//
// Copyright (c) 1999, Bernhard Tschirren
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of the source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//
// 2. Any redistribution solely in binary form must conspicuously
//    reproduce the following disclaimer in documentation provided with
//    the binary redistribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'', WITHOUT ANY WARRANTIES, EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  LICENSOR SHALL
// NOT BE LIABLE FOR ANY LOSS OR DAMAGES RESULTING FROM THE USE OF THIS
// SOFTWARE, EITHER ALONE OR IN COMBINATION WITH ANY OTHER SOFTWARE.
//

#include <mx/mx.h>
#include "util/BaseClass.h"
#include "util/Debug.h"
#include "util/Globals.h"
#include "core/Scene.h"
#include "parsers/Parser.h"
#include "nodes/PointLightNode.h"
#include "devices/OpenglRenderDevice.h"
#include "render/PortalRenderProcess.h"
#include "application/Intersection.h"

#pragma comment (lib, "comctl32.lib")  /* link with Common Controls lib  */
#pragma comment (lib, "opengl32.lib")  /* link with Microsoft OpenGL lib */
#pragma comment (lib, "glu32.lib")     /* link with Microsoft OpenGL lib */

#if defined(_DEBUG)
#  pragma comment (lib, "mxtkd.lib")   /* link with mxTK (Debug)         */
#else
#  pragma comment (lib, "mxtk.lib")    /* link with mxTK                 */
#endif


#define APPLICATION_TITLE     "Architectural Walkthrough Project v1.00"
#define APPLICATION_COPYRIGHT "Copyright (C) 1999, Bernhard Tschirren"
#define APPLICATION_WIDTH     800
#define APPLICATION_HEIGHT    600

#define RENDER_FOVY           60.0
#define RENDER_ASPECT         1.0
#define RENDER_ZNEAR          1.0
#define RENDER_ZFAR           1000.0

#define VIEWER_NODE           "viewer"
#define VIEWER_VELOCITY		  5.0
#define VIEWER_LIGHT          FALSE


static RenderDevice*  gDevice      = NULL;
static RenderProcess* gProcess     = NULL;
static Scene*         gScene       = NULL;

static RenderNode*    gViewer         = NULL;
static Point          gViewerPos      = Point(0,0,0);
static Float          gViewerAngle    = 0;
static Float          gViewerVelocity = 0;

static Boolean        gLighting    = TRUE;
static Boolean        gWireframe   = FALSE;


class MainWindow : public mxWindow
  {
  private: enum Events
             {
             ID_FILE_OPEN      = 1001,
             ID_FILE_EXIT      = 1002,

             ID_VIEW_WIREFRAME = 1101,
             ID_VIEW_LIGHTING  = 1102,
             ID_VIEW_OPTIONS   = 1103,

             ID_HELP_HOMEPAGE  = 1901,
             ID_HELP_ABOUT     = 1902
             };

  // Components
  private: mxMenuBar* _menuBar;

  public: MainWindow()
            : mxWindow(0, 0, 0, 0, 0, APPLICATION_TITLE, mxWindow::Normal)
            {
            // Create File Menu
            mxMenu* menuFile = new mxMenu();
            menuFile->add("&Open...", ID_FILE_OPEN);
            menuFile->addSeparator();
            menuFile->add("E&xit", ID_FILE_EXIT);

            // Create Options Menu
            mxMenu* menuView = new mxMenu();
            menuView->add("&Wireframe",  ID_VIEW_WIREFRAME);
            menuView->add("&Lighting",   ID_VIEW_LIGHTING);
            menuView->addSeparator();
            menuView->add("&Options...", ID_VIEW_OPTIONS);

            // Create Help Menu
            mxMenu* menuHelp = new mxMenu();
            menuHelp->add("&Home Page", ID_HELP_HOMEPAGE);
            menuHelp->addSeparator();
            menuHelp->add("&About", ID_HELP_ABOUT);

            // Create Menu Bar
            _menuBar = new mxMenuBar(this);
            _menuBar->addMenu("&File", menuFile);
            _menuBar->addMenu("&View", menuView);
            _menuBar->addMenu("&Help", menuHelp);

            // Set the initial check states
            _menuBar->setChecked(ID_VIEW_WIREFRAME, gWireframe);
            _menuBar->setChecked(ID_VIEW_LIGHTING,  gLighting);

            // Create Render Process
            gProcess = new PortalRenderProcess(Globals::getValuef("Render.fovY",   RENDER_FOVY),
                                               Globals::getValuef("Render.aspect", RENDER_ASPECT),
                                               Globals::getValuef("Render.zNear",  RENDER_ZNEAR),
                                               Globals::getValuef("Render.zFar",   RENDER_ZFAR));

            // Create OpenGL window
            gDevice = new OpenglRenderDevice(this, 0, 0, 0, 0, NULL, mxWindow::Normal);
            gDevice->setRenderProcess(gProcess);
            gDevice->initLights(gLighting, Color(0.3, 0.3, 0.3));
            gDevice->setRenderMode(gWireframe ? RenderDevice::RENDER_LINES : RenderDevice::RENDER_FILLED);

            // Make sure that we are drawn            
            setMenuBar(_menuBar);
            setBounds(20, 20,
                      Globals::getValuei("Application.width",  APPLICATION_WIDTH),
                      Globals::getValuei("Application.height", APPLICATION_HEIGHT));
            setVisible(TRUE);
            }

  public: virtual ~MainWindow()
            {
            }

  public: virtual Void redraw()
            {
            // @todo We should use proper semaphores here!
            //       This is important, because the Win32
            //       callback function is re-entrant and
            //       all our own code is not!
            //
            //       This bug took me AGES to find :)
            //
            static Boolean semaphore = FALSE;

            if (!semaphore)
              {
              semaphore = TRUE;
              gDevice->redraw();
              semaphore = FALSE;
              }
            }

  public: virtual Int handleEvent(mxEvent* event)
            {
            switch (event->event)
              {
              case mxEvent::Action:
                {
		        switch (event->action)
                  {
                  case ID_FILE_OPEN:
                    openScene();
                    break;

                  case ID_FILE_EXIT:
                    mx::quit();
                    break;

                  case ID_VIEW_LIGHTING:
                    gLighting = !gLighting;
                    _menuBar->setChecked(ID_VIEW_LIGHTING,  gLighting);
                    gDevice->initLights(gLighting, Color(0.3, 0.3, 0.3));
                    redraw();
                    break;

                  case ID_VIEW_WIREFRAME:
                    gWireframe = !gWireframe;
                    _menuBar->setChecked(ID_VIEW_WIREFRAME, gWireframe);
                    gDevice->setRenderMode(gWireframe ? RenderDevice::RENDER_LINES : RenderDevice::RENDER_FILLED);
                    redraw();
                    break;

                  case ID_VIEW_OPTIONS:
                    mxMessageBox (this,
                      "Take a look in 'project.ini' to see some more interresting view options.\n"
                      "You will need to restart the program before any changes take effect.\n",
                      "View Options",
                      MX_MB_OK | MX_MB_INFORMATION);
                    break;

                  case ID_HELP_HOMEPAGE:
                    // @todo Win32 Specific
                    ShellExecute(0, "open", "http://www.computing.edu.au/~tschirre/pub/project/", 0, 0, SW_SHOW);
                    break;

                  case ID_HELP_ABOUT:
                    mxMessageBox (this,
                      APPLICATION_TITLE "\n"
                      APPLICATION_COPYRIGHT "\n"
                      "\n"
                      "Build:\t" __DATE__ "\n"
                      "Email:\t" "tschirre@computing.edu.au" "\n"
                      "Web:\t"   "http://www.computing.edu.au/~tschirre/pub/project/" "\n",
                      "About",
                      MX_MB_OK | MX_MB_INFORMATION);
                    break;
                  }
                }
                break;
              
              case mxEvent::Size:
                {
		        Int w = event->width;
		        Int h = event->height;
		        Int y = _menuBar->getHeight();
		        gDevice->setBounds(0, y, w, h);
                }
                break;

              case mxEvent::KeyDown:
                {
                switch (event->key)
                  {
                  case 37:
                    gViewerAngle += 5 * M_PI/180;
                    break;

                  case 39:
                    gViewerAngle -= 5 * M_PI/180;
                    break;

                  case 38:
					gViewerPos += Vector(gViewerVelocity*cos(gViewerAngle),
						                 gViewerVelocity*sin(gViewerAngle),
						                 0);
                    break;

                  case 40:
					gViewerPos -= Vector(gViewerVelocity*cos(gViewerAngle),
					                     gViewerVelocity*sin(gViewerAngle),
						                 0);
                    break;

				  case 33:
					gViewerPos += Vector(0, 0, gViewerVelocity);
                    break;

                  case 34:
					gViewerPos -= Vector(0, 0, gViewerVelocity);
                    break;
                  }

                if (gViewer)
                  {
                  // Save previous position
                  Point prevPos = gViewer->getTranslation();

                  // Update current position and direction
                  gViewer->setTranslation(gViewerPos);
                  gViewer->setRotation(Quaternion().fromEuler(0,
                                                              90*M_PI/180 - gViewerAngle,
                                                              -90*M_PI/180));

                  // Check portal intersections
				  checkPortalIntersections(gViewer, prevPos);

                  // And finally, redraw
				  redraw();
                  }
                }
                break;
              }

            return 1;
            }

  private: Void openScene()
             {
             const Char* name = mxGetOpenFileName(this, ".", "*.sce");
             if (!name)
               return;

             mxWindow* window = new mxWindow(NULL, 0, 0, 0, 0, "Loading...", mxWindow::Dialog);
             window->setVisible(TRUE);
             try
               {
               // Create parser
               FileStream stream(name, "r");
               Parser     parser(&stream);

               // Parser file
               parser.parse();

               // Create a new scene
               if (gScene) delete gScene;
               gScene = new Scene(parser.getSymbols());

               // Locate viewer
               gViewer = (RenderNode*)parser.find(Globals::getValues("Viewer.node", VIEWER_NODE),
                                                  STATIC_TYPE_INFO(RenderNode));
               if (!gViewer)
                 mxMessageBox(this,
                              String().setFormatted("Scene does not contain viewer node: \"%s", Globals::getValues("Viewer.node", VIEWER_NODE)),
                              "Could Not Open Scene",
                              MX_MB_OK | MX_MB_ERROR);

               // Attach a point light source to viewer
               if (Globals::getValueb("Viewer.light", VIEWER_LIGHT))
                 {
                 RenderNode* node = new PointLightNode(Color(0.4, 0.4, 0.4),
                                                       Color(0.7, 0.7, 0.7),
                                                       Color(0.0, 0.0, 0.0),
                                                       1, 0.005, 0);
                 gViewer->attachChild(node);
                 gScene->add(node);
                 }

               // Retain viewer position
               gViewerPos      = gViewer->getTranslation();
               gViewerAngle    = 0;
               gViewerVelocity = Globals::getValuef("Viewer.velocity", VIEWER_VELOCITY);

               // Start rendering from given viewer
               gProcess->setViewer(gViewer);
               }
             catch (Exception e)
               {
               mxMessageBox(this, e.getMessage(), "Could Not Open Scene", MX_MB_OK | MX_MB_ERROR);
               }
             delete window;
             }
  };


int main(int argc, char* argv[])
  {
  Int ret = 0;

  mx::init(argc, argv);

  try
    {
    Globals::load(FileStream("project.ini", "r"));

    mxWindow* gMainWindow = new MainWindow();
    mx::run();
    }
  catch(Exception e)
    {
    // Write the exception into a file
    Debug debug("debug-Exceptions.txt");
    debug.write("Fatal Exception: %s\n", e.getMessage());

    // Display it on screen (this sometimes does not work!)
    mxMessageBox(NULL, e.getMessage(), "Fatal Exception", MX_MB_OK | MX_MB_ERROR);    

    // Signal to parent process that something went wrong
    ret = 1;
    }

  mx::cleanup();

  return ret;
  }