//
// Copyright (c) 1999, Bernhard Tschirren
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of the source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//
// 2. Any redistribution solely in binary form must conspicuously
//    reproduce the following disclaimer in documentation provided with
//    the binary redistribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'', WITHOUT ANY WARRANTIES, EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  LICENSOR SHALL
// NOT BE LIABLE FOR ANY LOSS OR DAMAGES RESULTING FROM THE USE OF THIS
// SOFTWARE, EITHER ALONE OR IN COMBINATION WITH ANY OTHER SOFTWARE.
//

#ifndef _BT_QUATERNION_H_
#define _BT_QUATERNION_H_

#include "util/TTuple.h"
#include "core/Vector.h"
#include "core/Matrix.h"

class Quaternion : public Tuple4
  {
  public: Quaternion()
            : Tuple4()
            {}

  public: Quaternion(const Quaternion& other)
            {
            set(other);
            }

  public: Quaternion(Float w, Float x, Float y, Float z)
            {
            set(w, x, y, z);
            }

  public: Float getW() const
            {
            return Tuple4::get(0);
            }

  public: Float getX() const
            {
            return Tuple4::get(1);
            }

  public: Float getY() const
            {
            return Tuple4::get(2);
            }

  public: Float getZ() const
            {
            return Tuple4::get(3);
            }

  public: Float lengthSquared() const
            {
            return Tuple4::lengthSquared();
            }

  public: Float length() const
            {
            return Tuple4::length();
            }

  public: Float equals(const Quaternion& other) const
            {
            return Tuple4::equals(other);
            }

  public: Float dotProduct(const Quaternion& other) const
            {
            return Tuple4::dotProduct(other);
            }

  public: Quaternion& set(Float w, Float x, Float y, Float z)
            {
            Tuple4::set(0, w);
            Tuple4::set(1, x);
            Tuple4::set(2, y);
            Tuple4::set(3, z);
            return *this;
            }

  public: Quaternion& set(const Quaternion& other)
            {
            return set(other.getW(),
                       other.getX(),
                       other.getY(),
                       other.getZ());
            }

  public: Quaternion& fromEuler(Float yaw, Float pitch, Float roll)
            {
            Float cr   = cos(roll/2);
            Float cp   = cos(pitch/2);
            Float cy   = cos(yaw/2);
            Float sr   = sin(roll/2);
            Float sp   = sin(pitch/2);
            Float sy   = sin(yaw/2);
            Float cpcy = cp * cy;
            Float cpsy = cp * sy;
            Float spsy = sp * sy;
            Float spcy = sp * cy;

            return set(cr * cpcy + sr * spsy,
                       sr * cpcy - cr * spsy,
                       cr * spcy + sr * cpsy,
                       cr * cpsy - sr * spcy);
            }

  public: Quaternion& fromAngleAxis(Float angle, const Vector& axis)
            {
            Float halfAngle = 0.5 * angle;
            Float sn        = sin(halfAngle);

            return set(cos(halfAngle),
                       sn * axis.getI(),
                       sn * axis.getJ(),
                       sn * axis.getK());
            }

  public: Void toAngleAxis(Float& angle, Vector& axis) const
            {
            Float length2 = getX() * getX() +
                            getY() * getY() +
                            getZ() * getZ();

            if (length2 > 0.0)
              {
              Float invLen = 1.0 / sqrt(length2);

              angle = 2.0 * acos(getW());
              axis.set(getX() * invLen,
                       getY() * invLen,
                       getZ() * invLen);
              }
            else
              {
              angle = 0;
              axis.set(1, 0, 0);
              }
            }

  public: Void toMatrix(Matrix& matrix) const
            {
            Float x2 = getX() + getX();
            Float y2 = getY() + getY();
            Float z2 = getZ() + getZ();
            Float xx = getX() * x2;
            Float xy = getX() * y2;
            Float xz = getX() * z2;
            Float yy = getY() * y2;
            Float yz = getY() * z2;
            Float zz = getZ() * z2;
            Float wx = getW() * x2;
            Float wy = getW() * y2;
            Float wz = getW() * z2;

            matrix.set(0,0, 1.0 - (yy + zz));
            matrix.set(0,1, xy - wz);
            matrix.set(0,2, xz + wy);
            matrix.set(0,3, 0.0);

            matrix.set(1,0, xy + wz);
            matrix.set(1,1, 1.0 - (xx + zz));
            matrix.set(1,2, yz - wx);
            matrix.set(1,3, 0.0);

            matrix.set(2,0, xz - wy);
            matrix.set(2,1, yz + wx);
            matrix.set(2,2, 1.0 - (xx + yy));
            matrix.set(2,3, 0.0);

            matrix.set(3,0, 0.0);
            matrix.set(3,1, 0.0);
            matrix.set(3,2, 0.0);
            matrix.set(3,3, 1.0);
            }

  public: Quaternion& interpolate(const Quaternion& other, Float time)
            {
            Quaternion        temp;
            const Quaternion* to;
            Float             cosom;
            Float             scale0;
            Float             scale1;

            // Calc cosine
            cosom = dotProduct(other);

            // Adjust signs (if necessary)
            if (cosom < 0.0)
              {
              cosom = -cosom;
              to    = &temp;

              temp.set(-other.getW(),
                       -other.getX(),
                       -other.getY(),
                       -other.getZ());             
              }
            else
              to = &other;

            // Clamp time to range [0,1]
            time = FLOAT_CLAMP(time, 0, 1);

            // Calculate coefficients
            if ((1.0 - cosom) > 0.01)
              {
              Float omega = acos(cosom);
              Float sinom = 1.0 / sin(omega);
              scale0 = sin((1.0 - time) * omega) * sinom;
              scale1 = sin(       time  * omega) * sinom;
              }
            else
              {
              scale0 = 1.0 - time;
              scale1 = time;
              }

            set(scale0 * getW() + scale1 * to->getW(),
                scale0 * getX() + scale1 * to->getX(),
                scale0 * getY() + scale1 * to->getY(),
                scale0 * getZ() + scale1 * to->getZ());
            }

  public: const Quaternion& operator =(const Quaternion& other)
            {
            set(other);
            return *this;
            }

  public: Quaternion operator +(const Quaternion& other) const
            {
            return Quaternion(getW() + other.getW(),
                              getX() + other.getX(),
                              getY() + other.getY(),
                              getZ() + other.getZ());
            }

  public: const Quaternion& operator +=(const Quaternion other)
            {
            set(getW() + other.getW(),
                getX() + other.getX(),
                getY() + other.getY(),
                getZ() + other.getZ());
            }

  public: Quaternion operator -(const Quaternion& other) const
            {
            return Quaternion(getW() - other.getW(),
                              getX() - other.getX(),
                              getY() - other.getY(),
                              getZ() - other.getZ());
            }

  public: const Quaternion& operator -=(const Quaternion other)
            {
            set(getW() - other.getW(),
                getX() - other.getX(),
                getY() - other.getY(),
                getZ() - other.getZ());
            }

  public: Quaternion operator *(const Quaternion& other) const
            {
            return Quaternion(getW()*other.getW() - getX()*other.getX() - getY()*other.getY() - getZ()*other.getZ(),
                              getW()*other.getX() + getX()*other.getW() + getY()*other.getZ() - getZ()*other.getY(),
                              getW()*other.getY() + getY()*other.getW() + getZ()*other.getX() - getX()*other.getZ(),
                              getW()*other.getZ() + getZ()*other.getW() + getX()*other.getY() - getY()*other.getX());
            }

  public: Quaternion operator -() const
            {
            return Quaternion(-getW(), -getX(), -getY(), -getZ());
            }

  TYPE_DATA // RTTI
  };

#endif /* _BT_QUATERNION_H_ */