//
// Copyright (c) 1999, Bernhard Tschirren
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of the source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//
// 2. Any redistribution solely in binary form must conspicuously
//    reproduce the following disclaimer in documentation provided with
//    the binary redistribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'', WITHOUT ANY WARRANTIES, EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  LICENSOR SHALL
// NOT BE LIABLE FOR ANY LOSS OR DAMAGES RESULTING FROM THE USE OF THIS
// SOFTWARE, EITHER ALONE OR IN COMBINATION WITH ANY OTHER SOFTWARE.
//

#ifndef _BT_RENDERDEVICE_H_
#define _BT_RENDERDEVICE_H_

#include <mx/mx.h>
#include "util/BaseClass.h"
#include "core/Point.h"
#include "core/Vector.h"
#include "core/Color.h"
#include "core/Quaternion.h"
#include "core/Matrix.h"
#include "core/Geometry.h"
#include "core/RenderProcess.h"

class RenderDevice : public BaseClass, public mxWindow
  {
  public: enum ClearMode
            {
            CLEAR_COLOR_BUFFER = 0x01,
            CLEAR_DEPTH_BUFFER = 0x02
            };

  public: enum ObjectMode
            {
            RENDER_OBJECT  = 0x01,
            COMPILE_OBJECT = 0x02
            };

  public: enum RenderMode
            {
            RENDER_POINTS,
            RENDER_LINES,
            RENDER_FILLED
            };

  private: RenderProcess* _renderProcess;

  protected: RenderDevice(mxWindow *parent, Int x, Int y, Int w, Int h, const Char* label, Int style)
               : mxWindow(parent, x, y, w, h, label, style),
                 _renderProcess(NULL)
               {
               }

  public: Void setRenderProcess(RenderProcess* renderProcess)
            {
            _renderProcess = renderProcess;
            }

  public: Void redraw()
            {
            if (_renderProcess)
              _renderProcess->render(this);
            }

  public: virtual Void setPerspective (Float left, Float right, Float bottom, Float top, Float zNear, Float zFar) = 0;
  public: virtual Void setOrthographic(Float left, Float right, Float bottom, Float top, Float zNear, Float zFar) = 0;

  public: virtual Void setRenderMode(RenderMode mode) = 0;

  public: virtual Void clear(Int mode) = 0;

  public: virtual Void beginScene() = 0;
  public: virtual Void endScene()   = 0;

  public: virtual Int  beginObject(Int mode) = 0;
  public: virtual Void endObject(Int obj)    = 0;
  public: virtual Void deleteObject(Int obj) = 0;

  public: virtual Void scale(const Vector& scale)          = 0;
  public: virtual Void translate(const Point& translation) = 0;
  public: virtual Void rotate(const Quaternion& rotation)  = 0;

  public: virtual Void   clearMatrix()                   = 0;
  public: virtual Void   setMatrix(const Matrix& matrix) = 0;
  public: virtual Matrix getMatrix()                     = 0;
  public: virtual Void   pushMatrix()                    = 0;
  public: virtual Void   popMatrix()                     = 0;

  public: virtual Void initLights         (const Boolean enabled,
                                           const Color& ambient) = 0;
  public: virtual Void addPointLight      (const Color& ambient,
                                           const Color& diffuse,
                                           const Color& specular,
                                           const Float  constantAttenuation,
                                           const Float  linearAttenuation,
                                           const Float  quadraticAttenuation) = 0;
  public: virtual Void addDirectionalLight(const Color& ambient,
                                           const Color& diffuse,
                                           const Color& specular,
                                           const Float  constantAttenuation,
                                           const Float  linearAttenuation,
                                           const Float  quadraticAttenuation,
                                           const Float  exponent,
                                           const Float  cutoff) = 0;
  public: virtual Void addPlanarLight     (const Color& ambient,
                                           const Color& diffuse,
                                           const Color& specular) = 0;
  public: virtual Void clearLights() = 0;
  public: virtual Void pushLights()  = 0;
  public: virtual Void popLights()   = 0;

  public: virtual Void render(Int obj)                  = 0;
  public: virtual Void render(const Geometry& geometry) = 0;

  TYPE_DATA // RTTI
  };

class RenderDeviceException : public Exception
  {
  public: RenderDeviceException(const RenderDevice* device, const Char* format, ...)
            {
            va_list args;

            va_start(args, format);
            vsprintf(_message, format, args);
            va_end(args);
            };
  };

#endif /* _BT_RENDERDEVICE_H_ */