//
// Copyright (c) 1999, Bernhard Tschirren
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of the source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//
// 2. Any redistribution solely in binary form must conspicuously
//    reproduce the following disclaimer in documentation provided with
//    the binary redistribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'', WITHOUT ANY WARRANTIES, EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  LICENSOR SHALL
// NOT BE LIABLE FOR ANY LOSS OR DAMAGES RESULTING FROM THE USE OF THIS
// SOFTWARE, EITHER ALONE OR IN COMBINATION WITH ANY OTHER SOFTWARE.
//

#ifndef _BT_VECTOR_H_
#define _BT_VECTOR_H_

#include "util/TTuple.h"

class Vector : public Tuple3
  {
  public: Vector()
            {
            }

  public: Vector(Float i, Float j, Float k)
            {
            Vector::set(i, j, k);
            }

  public: Vector(const Vector& other)
            {
            Vector::set(other);
            }

  public: Float getI() const
            {
            return Tuple3::get(0);
            }

  public: Float& getI()
            {
            return Tuple3::getRef(0);
            }

  public: Float getJ() const
            {
            return Tuple3::get(1);
            }

  public: Float& getJ()
            {
            return Tuple3::getRef(1);
            }

  public: Float getK() const
            {
            return Tuple3::get(2);
            }

  public: Float& getK()
            {
            return Tuple3::getRef(2);
            }

  public: Vector& set(Float i, Float j, Float k)
            {
            Tuple3::set(0, i);
            Tuple3::set(1, j);
            Tuple3::set(2, k);
            return *this;
            }

  public: Vector& set(const Vector& other)
            {
            Tuple3::set(other);
            return *this;
            }

  public: Boolean equals(const Vector& other) const
            {
            return Tuple3::equals(other);
            }

  public: Boolean isZero() const
            {
            return Tuple3::isZero();
            }

  public: Float lengthSquared() const
            {
            return Tuple3::lengthSquared();
            }

  public: Float length() const
            {
            return Tuple3::length();
            }

  public: Vector& scale(Float factor)
            {
            Tuple3::scale(factor);
            return *this;
            }

  public: Vector& scale(const Vector& factor)
            {
            Tuple3::scale(factor);
            return *this;
            }

  public: Vector& normalise()
            {
            Float length2 = lengthSquared();

            ASSERT(!FLOAT_IS_ZERO(length2));

            if (FLOAT_EQUALS(length2, 1.0))
              return *this;
            else
              return scale((Float)1.0 / (Float)sqrt(length2));
            }

  public: Float dotProduct(const Vector& other) const
            {
            return Tuple3::dotProduct(other);
            }

  public: Vector crossProduct(const Vector& other) const
            {
            return Vector((getJ() * other.getK()) - (getK() * other.getJ()),
                          (getK() * other.getI()) - (getI() * other.getK()),
                          (getI() * other.getJ()) - (getJ() * other.getI()));
            }

  public: Vector& interpolate(const Vector& other, Float time)
            {
            Tuple3::interpolate(other, time);
            return *this;
            }

  public: const Vector& operator =(const Vector& other)
            {
            set(other);
            return *this;
            }

  public: Boolean operator ==(const Vector& other) const
            {
            return equals(other);
            }

  public: Boolean operator !=(const Vector& other) const
            {
            return !equals(other);
            }

  public: Vector& operator *=(Float factor)
            {
            return scale(factor);
            }

  public: Vector& operator *(Float factor)
            {
            return scale(factor);
            }

  public: Vector& operator /=(Float factor)
            {
            ASSERT(!FLOAT_IS_ZERO(factor));
            return scale((Float)1.0 / factor);
            }

  public: Vector& operator /(Float factor)
            {
            ASSERT(!FLOAT_IS_ZERO(factor));
            return scale((Float)1.0 / factor);
            }

  public: friend Vector operator +(const Vector& v1, const Vector& v2)
            {
            return Vector(v1.getI() + v2.getI(),
                          v1.getJ() + v2.getJ(),
                          v1.getK() + v2.getK());
            }

  public: const Vector& operator +=(const Vector& v)
            {
            getI() += v.getI();
            getJ() += v.getJ();
            getK() += v.getK();
            return *this;
            }

  public: friend Vector operator -(const Vector& v1, const Vector& v2)
            {
            return Vector(v1.getI() - v2.getI(),
                          v1.getJ() - v2.getJ(),
                          v1.getK() - v2.getK());
            }

  public: const Vector& operator -=(const Vector& v)
            {
            getI() -= v.getI();
            getJ() -= v.getJ();
            getK() -= v.getK();
            return *this;
            }

  public: friend Vector operator -(const Vector& v)
            {
            return Vector(-v.getI(), -v.getJ(), -v.getK());
            }

  TYPE_DATA // RTTI
  };

#endif /* _BT_VECTOR_H_ */