//
// Copyright (c) 1999, Bernhard Tschirren
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of the source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//
// 2. Any redistribution solely in binary form must conspicuously
//    reproduce the following disclaimer in documentation provided with
//    the binary redistribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'', WITHOUT ANY WARRANTIES, EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  LICENSOR SHALL
// NOT BE LIABLE FOR ANY LOSS OR DAMAGES RESULTING FROM THE USE OF THIS
// SOFTWARE, EITHER ALONE OR IN COMBINATION WITH ANY OTHER SOFTWARE.
//

#include "devices/OpenglRenderDevice.h"
#include "util/Globals.h"
#include "util/Debug.h"
#include <GL/gl.h>
#include <GL/glext.h>
#include <GL/glu.h>
#include <mx/mxTga.h>

#ifndef GL_VERSION_1_1
#  error "You need at least OpenGL 1.1 headers!"
#endif

#define USE_DYNAMIC_LINKING
#define USE_ERROR_CHECKING


static PFNGLLOCKARRAYSEXTPROC          glLockArraysEXT;
static PFNGLUNLOCKARRAYSEXTPROC        glUnlockArraysEXT;
static PFNGLBEGINSCENEEXTPROC          glBeginSceneEXT;
static PFNGLENDSCENEEXTPROC            glEndSceneEXT;
static PFNGLACTIVETEXTUREARBPROC       glActiveTextureARB;
static PFNGLCLIENTACTIVETEXTUREARBPROC glClientActiveTextureARB;


// Create a debug object
static Debug debug("debug-OpenglRenderDevice.txt");

// Setup some more debugging stuff. The do/while(0) loop
// is used to turn the definition into a single C statement
// which needs a semi-colon terminator.
#ifdef USE_ERROR_CHECKING
  #define GL_CHECK(function)                            \
    do                                                  \
      {                                                 \
      function;                                         \
                                                        \
      Int error = glGetError();                         \
      if (error != GL_NO_ERROR)                         \
        debug.write("  %s:%d: %s: GL_ERROR: %s\n",      \
                    __FILE__, __LINE__,                 \
                    #function, gluErrorString(error));  \
      }                                                 \
    while(0)

#else /* ! USE_ERROR_CHECKING */
  #define GL_CHECK(function)                            \
    function
#endif /* USE_ERROR_CHECKING */


OpenglRenderDevice::OpenglRenderDevice(mxWindow* parent, Int x, Int y, Int w, Int h, const Char* label, Int style)
  : RenderDevice(parent, x, y, w, h, label, style),
    _opengl(NULL),
    _lights(),
    _numTextureUnits(0),
    _hdc(NULL),
    _hglrc(NULL)
  {
  debug.write("Constructing OpenglRenderDevice\n");

  // Make sure datatypes match
  ASSERT_ALWAYS(sizeof(GLint)   == sizeof(Int));
  ASSERT_ALWAYS(sizeof(GLfloat) == sizeof(Float));

#ifdef USE_DYNAMIC_LINKING
  // Load OpenGL DLL
  _opengl = new Library("opengl32");
#endif /* USE_DYNAMIC_LINKING */

  // Create OpenGL context
  newContext();

  // Dynamically link extensions
  glLockArraysEXT          = (PFNGLLOCKARRAYSEXTPROC)          link("glLockArraysEXT",          "GL_EXT_compiled_vertex_array", TRUE);
  glUnlockArraysEXT        = (PFNGLUNLOCKARRAYSEXTPROC)        link("glUnlockArraysEXT",        "GL_EXT_compiled_vertex_array", TRUE);
  glBeginSceneEXT          = (PFNGLBEGINSCENEEXTPROC)          link("glBeginSceneEXT",          "GL_EXT_scene_marker",          TRUE);
  glEndSceneEXT            = (PFNGLENDSCENEEXTPROC)            link("glEndSceneEXT",            "GL_EXT_scene_marker",          TRUE);
  glActiveTextureARB       = (PFNGLACTIVETEXTUREARBPROC)       link("glActiveTextureARB",       "GL_ARB_multitexture",          TRUE);
  glClientActiveTextureARB = (PFNGLCLIENTACTIVETEXTUREARBPROC) link("glClientActiveTextureARB", "GL_ARB_multitexture",          TRUE);

  if (Globals::getValueb("OpenglRenderDevice.texturing", TRUE))
    {
    if (Globals::getValueb("OpenglRenderDevice.multitexturing", TRUE) && glActiveTextureARB && glClientActiveTextureARB)
      {
      // Single-texturing enabled, multi-texturing enabled
      glGetIntegerv(GL_MAX_TEXTURE_UNITS_ARB, &_numTextureUnits);
      }
    else
      {
      // Single-texturing enabled, multi-texturing disabled
      glActiveTextureARB       = NULL;
      glClientActiveTextureARB = NULL;
      _numTextureUnits         = 1;
      }
    }
  else
    {
    // Single-texturing disable, multi-texturing disabled
    glActiveTextureARB       = NULL;
    glClientActiveTextureARB = NULL;
    _numTextureUnits         = 0;
    }

  // Write some debug info into file
  debug.write("  GL_VENDOR:                %s\n", glGetString(GL_VENDOR));
  debug.write("  GL_RENDERER:              %s\n", glGetString(GL_RENDERER));
  debug.write("  GL_VERSION:               %s\n", glGetString(GL_VERSION));
  debug.write("  GL_EXTENSIONS:            %s\n", glGetString(GL_EXTENSIONS));
  debug.write("  GL_MAX_TEXTURE_UNITS_ARB: %d\n", _numTextureUnits);

  // Setup initial OpenGL state
  GL_CHECK(glEnableClientState(GL_VERTEX_ARRAY));
  GL_CHECK(glEnableClientState(GL_NORMAL_ARRAY));
  GL_CHECK(glEnable(GL_DEPTH_TEST));
  GL_CHECK(glEnable(GL_CULL_FACE));
  GL_CHECK(glEnable(GL_DITHER));
  GL_CHECK(glEnable(GL_BLEND));  
  GL_CHECK(glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA));
  GL_CHECK(glDisable(GL_COLOR_MATERIAL));
  GL_CHECK(glDisable(GL_LIGHTING));
  GL_CHECK(glDisable(GL_TEXTURE_1D));
  GL_CHECK(glDisable(GL_TEXTURE_2D));
  GL_CHECK(glShadeModel(GL_SMOOTH));
  GL_CHECK(glPolygonMode(GL_FRONT_AND_BACK, GL_FILL));

  // Make sure we have a current light list
  _lights.setLength(1);

  // Some MX windowing stuff
  setType(MX_GLWINDOW);
  parent->addWidget(this);
  }


OpenglRenderDevice::~OpenglRenderDevice()
  {
  debug.write("Destructing OpenglRenderDevice\n\n");

  deleteContext();

#ifdef USE_DYNAMIC_LINKING
  delete _opengl;
#endif /* USE_DYNAMIC_LINKING */
  }


Void OpenglRenderDevice::redraw()
  {
  // Set this OpenGL context as the current one
  wglMakeCurrent(_hdc, _hglrc);

  // Call default redraw() method
  RenderDevice::redraw();

  SwapBuffers(_hdc);
  }


Void OpenglRenderDevice::setPerspective(Float left, Float right, Float bottom, Float top, Float zNear, Float zFar)
  {
  GL_CHECK(glMatrixMode(GL_PROJECTION));
  GL_CHECK(glLoadIdentity());
  GL_CHECK(glViewport(0, 0, w2(), h2()));
  GL_CHECK(glFrustum(left, right, bottom, top, zNear, zFar));
  GL_CHECK(glMatrixMode(GL_MODELVIEW));
  }


Void OpenglRenderDevice::setOrthographic(Float left, Float right, Float bottom, Float top, Float zNear, Float zFar)
  {
  GL_CHECK(glMatrixMode(GL_PROJECTION));
  GL_CHECK(glLoadIdentity());
  GL_CHECK(glViewport(0, 0, w2(), h2()));
  GL_CHECK(glOrtho(left, right, bottom, top, zNear, zFar));
  GL_CHECK(glMatrixMode(GL_MODELVIEW));
  }


Void OpenglRenderDevice::setRenderMode(RenderMode mode)
  {
  switch (mode)
    {
    case RENDER_POINTS: GL_CHECK(glPolygonMode(GL_FRONT_AND_BACK, GL_POINT)); break;
    case RENDER_LINES:  GL_CHECK(glPolygonMode(GL_FRONT_AND_BACK, GL_LINE));  break;
    case RENDER_FILLED: GL_CHECK(glPolygonMode(GL_FRONT_AND_BACK, GL_FILL));  break;
    }
  }


Void OpenglRenderDevice::clear(Int mode)
  {
  GL_CHECK(glClear(((mode & CLEAR_COLOR_BUFFER) ? GL_COLOR_BUFFER_BIT : 0) |
                   ((mode & CLEAR_DEPTH_BUFFER) ? GL_DEPTH_BUFFER_BIT : 0)));
  }


Void OpenglRenderDevice::beginScene()
  {
  // Begin rendering
  if (glBeginSceneEXT) GL_CHECK(glBeginSceneEXT());

  // Initialise the matrix stack
  GL_CHECK(glMatrixMode(GL_MODELVIEW));
  GL_CHECK(glLoadIdentity());
  }


Void OpenglRenderDevice::endScene()
  {
  // End rendering
  if (glEndSceneEXT) GL_CHECK(glEndSceneEXT());
  }


Int OpenglRenderDevice::beginObject(Int mode)
  {
  Int list = -1;

  if (mode & COMPILE_OBJECT)
    {
    list = glGenLists(1);
    GL_CHECK(glNewList(list, (mode & RENDER_OBJECT) ? GL_COMPILE_AND_EXECUTE : GL_COMPILE));
    }

  return list;
  }


Void OpenglRenderDevice::endObject(Int obj)
  {
  if (obj != -1)
    GL_CHECK(glEndList());
  }


Void OpenglRenderDevice::deleteObject(Int obj)
  {
  if (obj != -1)
    GL_CHECK(glDeleteLists(obj, 1));
  }


Void OpenglRenderDevice::scale(const Vector& scale)
  {
  GL_CHECK(glScalef(scale.getI(), scale.getJ(), scale.getK()));
  }


Void OpenglRenderDevice::translate(const Point& translation)
  {
  GL_CHECK(glTranslatef(translation.getX(), translation.getY(), translation.getZ()));
  }


Void OpenglRenderDevice::rotate(const Quaternion& rotation)
  {
  Float  angle;
  Vector axis;

  rotation.toAngleAxis(angle, axis);

  GL_CHECK(glRotatef(angle * 180.0 / M_PI,
                     axis.getI(), axis.getJ(), axis.getK()));
  }


Void OpenglRenderDevice::clearMatrix()
  {
  GL_CHECK(glLoadIdentity());
  }


Void OpenglRenderDevice::setMatrix(const Matrix& matrix)
  {
  GL_CHECK(glLoadMatrixf(((Matrix&)matrix).getPtr()));
  }


Matrix OpenglRenderDevice::getMatrix()
  {
  Matrix matrix;

  GL_CHECK(glGetFloatv(GL_MODELVIEW_MATRIX, matrix.getPtr()));

  return matrix;
  }


Void OpenglRenderDevice::pushMatrix()
  {
  GL_CHECK(glPushMatrix());
  }


Void OpenglRenderDevice::popMatrix()
  {
  GL_CHECK(glPopMatrix());
  }


Void OpenglRenderDevice::initLights(const Boolean enabled, const Color&  ambient)
  {
  Boolean accurate = TRUE;

  if (enabled)
    {
    GL_CHECK(glEnable(GL_LIGHTING));
    GL_CHECK(glLightModeli (GL_LIGHT_MODEL_TWO_SIDE,     0));
    GL_CHECK(glLightModeli (GL_LIGHT_MODEL_LOCAL_VIEWER, accurate ? 1 : 0));
    GL_CHECK(glLightModelfv(GL_LIGHT_MODEL_AMBIENT,      ambient.getPtr()));
    }
  else
    GL_CHECK(glDisable(GL_LIGHTING));
  }


Void OpenglRenderDevice::setLight(Int depth, Int index)
  {
  Light&  light        = _lights[depth][index];
  GLint   lightName    = GL_LIGHT0 + index;
  GLfloat position[4]  = { 0, 0, 0, 1 };
  GLfloat direction[3] = { 0, 0, -1   };

  // Adjust position if light is planar
  if (light.type == PLANAR_LIGHT)
    position[3] = 0;

  // Set details
  GL_CHECK(glLightfv(lightName, GL_POSITION,              position));
  GL_CHECK(glLightfv(lightName, GL_SPOT_DIRECTION,        direction));
  GL_CHECK(glLightf (lightName, GL_SPOT_EXPONENT,         light.exponent));
  GL_CHECK(glLightf (lightName, GL_SPOT_CUTOFF,           light.cutoff));

  // Set color
  GL_CHECK(glLightfv(lightName, GL_AMBIENT,               light.ambient.getPtr()));
  GL_CHECK(glLightfv(lightName, GL_DIFFUSE,               light.diffuse.getPtr()));
  GL_CHECK(glLightfv(lightName, GL_SPECULAR,              light.specular.getPtr()));

  // Set attenuation
  GL_CHECK(glLightf (lightName, GL_CONSTANT_ATTENUATION,  light.constantAttenuation));
  GL_CHECK(glLightf (lightName, GL_LINEAR_ATTENUATION,    light.linearAttenuation));
  GL_CHECK(glLightf (lightName, GL_QUADRATIC_ATTENUATION, light.quadraticAttenuation));

  // Enable light
  GL_CHECK(glEnable (lightName));
  }


Void OpenglRenderDevice::addPointLight(const Color& ambient,
                                       const Color& diffuse,
                                       const Color& specular,
                                       const Float  constantAttenuation,
                                       const Float  linearAttenuation,
                                       const Float  quadraticAttenuation)
  {
  Int    depth = _lights.getLength()-1;
  Int    index = _lights[depth].getLength(); _lights[depth].setLength(index+1);
  Light& light = _lights[depth][index];

  light.type                 = POINT_LIGHT;
  light.ambient              = ambient;
  light.diffuse              = diffuse;
  light.specular             = specular;
  light.constantAttenuation  = constantAttenuation;
  light.linearAttenuation    = linearAttenuation;
  light.quadraticAttenuation = quadraticAttenuation;
  light.exponent             = 0;
  light.cutoff               = 180;
  light.matrix               = getMatrix();

  setLight(depth, index);
  }


Void OpenglRenderDevice::addDirectionalLight(const Color& ambient,
                                             const Color& diffuse,
                                             const Color& specular,
                                             const Float  constantAttenuation,
                                             const Float  linearAttenuation,
                                             const Float  quadraticAttenuation,
                                             const Float  exponent,
                                             const Float  cutoff)
  {
  Int    depth = _lights.getLength()-1;
  Int    index = _lights[depth].getLength(); _lights[depth].setLength(index+1);
  Light& light = _lights[depth][index];

  light.type                 = DIRECTIONAL_LIGHT;
  light.ambient              = ambient;
  light.diffuse              = diffuse;
  light.specular             = specular;
  light.constantAttenuation  = constantAttenuation;
  light.linearAttenuation    = linearAttenuation;
  light.quadraticAttenuation = quadraticAttenuation;
  light.exponent             = exponent;
  light.cutoff               = cutoff;
  light.matrix               = getMatrix();

  setLight(depth, index);
  }


Void OpenglRenderDevice::addPlanarLight(const Color& ambient,
                                        const Color& diffuse,
                                        const Color& specular)
  {
  Int    depth = _lights.getLength()-1;
  Int    index = _lights[depth].getLength(); _lights[depth].setLength(index+1);
  Light& light = _lights[depth][index];

  light.type                 = PLANAR_LIGHT;
  light.ambient              = ambient;
  light.diffuse              = diffuse;
  light.specular             = specular;
  light.constantAttenuation  = 0;
  light.linearAttenuation    = 0;
  light.quadraticAttenuation = 0;
  light.exponent             = 0;
  light.cutoff               = 180;
  light.matrix               = getMatrix();

  setLight(depth, index);
  }


Void OpenglRenderDevice::clearLights()
  {
  Int depth     = _lights.getLength()-1;
  Int numLights = _lights[depth].getLength();

  for (Int light = 0; light < numLights; light++)
    GL_CHECK(glDisable(GL_LIGHT0 + light));
  
  _lights[depth].clear();
  }


Void OpenglRenderDevice::pushLights()
  {
  Int depth = _lights.getLength()-1;

  // Push stack
  _lights.add(_lights[depth]);
  }


Void OpenglRenderDevice::popLights()
  {
  Int depth = _lights.getLength()-1;

  // Remove current lights
  clearLights();

  if (depth > 0)
    {
    // Pop stack
    _lights.setLength(depth);
    depth -= 1;

    // Save model-view matrix
    pushMatrix();

    // Set lights that are currently on top of the stack
    Int numLights  = _lights[depth].getLength();
    for (Int index = 0; index < numLights; index++)
      {
      setMatrix(_lights[depth][index].matrix);
      setLight(depth, index);
      }

    // Restore model-view matrix
    popMatrix();
    }
  }


Void OpenglRenderDevice::render(Int obj)
  {
  glCallList(obj);
  }


Void OpenglRenderDevice::render(const Geometry& geometry)
  {  
  const Geometry::CoordArray&   coords   = geometry.getCoords();
  const Geometry::NormalArray&  normals  = geometry.getNormals();
  const Geometry::IndexArray&   indices  = geometry.getIndices();
  const Geometry::SurfaceArray& surfaces = geometry.getSurfaces();
  const Material*               material = geometry.getMaterial();

  // Set vertex arrays
  GL_CHECK(glVertexPointer(3, GL_FLOAT, sizeof(Geometry::Coord),  coords.getPtr()));
  GL_CHECK(glNormalPointer(   GL_FLOAT, sizeof(Geometry::Normal), normals.getPtr()));

  // Compile vertex arrays
  if (glLockArraysEXT) GL_CHECK(glLockArraysEXT(0, coords.getLength()));

  // Set material (optional)
  if (material != NULL)
    {
    GL_CHECK(glMaterialfv(GL_FRONT, GL_AMBIENT,   material->getAmbient().getPtr()));
    GL_CHECK(glMaterialfv(GL_FRONT, GL_DIFFUSE,   material->getDiffuse().getPtr()));
    GL_CHECK(glMaterialfv(GL_FRONT, GL_SPECULAR,  material->getSpecular().getPtr()));
    GL_CHECK(glMaterialfv(GL_FRONT, GL_EMISSION,  material->getEmission().getPtr()));
    GL_CHECK(glMaterialf (GL_FRONT, GL_SHININESS, material->getShininess()));
    }

  // Ok, setup some variables
  Int numSurfaces = MIN(_numTextureUnits, surfaces.getLength());
  Int sIndex      = 0;

  // setup any used texture units
  for (; sIndex < numSurfaces; sIndex++)
    {
    const Surface& surface               = surfaces[sIndex];
    const Surface::CoordArray& texCoords = surface.getCoords();

    // Activate texure unit
    if (glActiveTextureARB)
      {
      GL_CHECK(glActiveTextureARB      (GL_TEXTURE0_ARB + sIndex));
      GL_CHECK(glClientActiveTextureARB(GL_TEXTURE0_ARB + sIndex));      
      }

    // Enable texturing
    GL_CHECK(glEnableClientState(GL_TEXTURE_COORD_ARRAY));
    GL_CHECK(glEnable(GL_TEXTURE_2D));

    // Setup texture matrix
    GL_CHECK(glMatrixMode(GL_TEXTURE));
    OpenglRenderDevice::clearMatrix();
    OpenglRenderDevice::scale    (surface.getScale());
    OpenglRenderDevice::translate(surface.getTranslation());
    OpenglRenderDevice::rotate   (surface.getRotation());
    GL_CHECK(glMatrixMode(GL_MODELVIEW));

    // Upload texture
    uploadTexture(surface.getTexture());
    GL_CHECK(glTexCoordPointer(2, GL_FLOAT, sizeof(Surface::Coord), texCoords.getPtr()));
    }

  // Disable any unused texture units
  for (; sIndex < _numTextureUnits; sIndex++)
    {
    if (glActiveTextureARB)
      {
      GL_CHECK(glActiveTextureARB      (GL_TEXTURE0_ARB + sIndex));
      GL_CHECK(glClientActiveTextureARB(GL_TEXTURE0_ARB + sIndex));
      }
    GL_CHECK(glDisableClientState(GL_TEXTURE_COORD_ARRAY));
    GL_CHECK(glDisable(GL_TEXTURE_2D));
    }

  // Draw triangles
  GL_CHECK(glDrawElements(GL_TRIANGLES, indices.getLength(), GL_UNSIGNED_INT, indices.getPtr()));

  // Unlock arrays
  if (glUnlockArraysEXT) GL_CHECK(glUnlockArraysEXT());
  }


Void OpenglRenderDevice::newContext()
  {
  PIXELFORMATDESCRIPTOR pfd =
    {
    sizeof (PIXELFORMATDESCRIPTOR),   // size of this pfd
    1,                                // version number
    PFD_DRAW_TO_WINDOW |              // support window
    PFD_SUPPORT_OPENGL |              // support OpenGL
    PFD_DOUBLEBUFFER,                 // double buffered
    PFD_TYPE_RGBA,                    // RGBA type
    32,                               // 32-bit color depth
    0, 0, 0, 0, 0, 0,                 // color bits ignored
    0,                                // no alpha buffer
    0,                                // shift bit ignored
    0,                                // no accumulation buffer
    0, 0, 0, 0,                       // accum bits ignored
    32,                               // 32-bit z-buffer
    0,                                // no stencil buffer
    0,                                // no auxiliary buffer
    PFD_MAIN_PLANE,                   // main layer
    0,                                // reserved
    0, 0, 0                           // layer masks ignored
    };
  Boolean ok;
  Int     pfm;

  _hdc = GetDC((HWND)getHandle());
  if (!_hdc)
    throw RenderDeviceException(this, "GetDC() failed [%d]", GetLastError());

  pfm = ChoosePixelFormat(_hdc, &pfd);
  if (!pfm)
    throw RenderDeviceException(this, "ChoosePixelFormat() failed [%d]", GetLastError());

  ok = (SetPixelFormat(_hdc, pfm, &pfd) != 0);
  if (!ok)
    throw RenderDeviceException(this, "SetPixelFormat() failed [%d]", GetLastError());

  _hglrc = wglCreateContext(_hdc);
  if (!_hglrc)
    throw RenderDeviceException(this, "wglCreateContext() failed [%d]", GetLastError());

  ok = (wglMakeCurrent(_hdc, _hglrc) != 0);
  if (!ok)
    throw RenderDeviceException(this, "wglMakeCurrent() failed [%d]", GetLastError());
  }


Void OpenglRenderDevice::deleteContext()
  {
  if (_hglrc)
    {
    wglMakeCurrent(NULL, NULL);
    wglDeleteContext(_hglrc);
    }
  
  if (_hdc)
    ReleaseDC((HWND)getHandle(), _hdc);
  }


Boolean OpenglRenderDevice::supportsExtension(const Char* extension_)
  {
  // Convert everything to (char*)
  static const char* extensions = NULL;
  const char*        extension  = extension_;
  const char*        start;
  char*              where;
  char*              terminator;
  Boolean            supported;

  // Extension names should not have spaces
  where = strchr(extension, ' ');
  if (where || *extension == '\0')
    return FALSE;

  // Grab extensions (but only once)
  if (!extensions)
    extensions = (const char*)glGetString(GL_EXTENSIONS);

  // Look for extension
  start     = extensions;
  supported = FALSE;
  while (!supported)
    {
    // Does extension SEEM to be supported?
    where = strstr((const char*)start, extension);
    if (!where)
      break;

    // Ok, extension SEEMS to be supported
    supported = TRUE;

    // Check for space before extension
    supported &= (where == start) || (where[-1] == ' ');

    // Check for space after extension
    terminator = where + strlen(extension);
    supported &= (*terminator == '\0') || (*terminator == ' ');

    // Next search starts at current terminator
    start = terminator;
    }

  return supported;
  }


Void* OpenglRenderDevice::link(const Char* identifier, const Char* extension, Boolean optional)
  {
  Void* address = NULL;

#ifdef USE_DYNAMIC_LINKING

  debug.write("  Linking %s():", identifier);
  
  // First try linking
  address = _opengl->link(identifier);

  if (address)
    debug.write(" Function found at address 0x%08x", address);
  else
    debug.write(" Function not found");
  
  // Check that extension is supported
  if ((address != NULL) && (extension[0] != '\0'))
    {
    if (!supportsExtension(extension))
      {
      debug.write(" but extension '%s' is not supported!", extension);
      address = NULL;
      }
    else
      debug.write(" and extension '%s' is supported", extension);
    }

  debug.write("\n");

#endif /* USE_DYNAMIC_LINKING */
  
  // Report any linking errors
  if (!address && !optional)
    throw RenderDeviceException(this, "Could not link \"%s\"!", (Char*)identifier);

  return address;
  }


Void OpenglRenderDevice::uploadTexture(Texture* texture)
  {
  if (texture->getHandle() != 0)
    GL_CHECK(glBindTexture(GL_TEXTURE_2D, texture->getHandle()));

  else
    {
    UInt    newHandle;
    Boolean mipmap;
    Int     minFilter;
    Int     magFilter;

    // Load image from file
    mxImage* image = mxTgaRead(texture->getFileName());
    if (image == NULL)
      throw RenderDeviceException(this, "Could not load \"%s\"!", (const Char*)texture->getFileName());

    // Create a new texture object
    GL_CHECK(glGenTextures(1, &newHandle));
    GL_CHECK(glBindTexture(GL_TEXTURE_2D, newHandle));
    texture->setHandle(newHandle);

    // Do we mipMap?
    if (texture->getMipMap())
      mipmap = Globals::getValueb("OpenglRenderDevice.mipmap", TRUE);
    else
      mipmap = FALSE;

    // Get minFilter
    if (Globals::getValueb("OpenglRenderDevice.minFilter", TRUE))
      minFilter = GL_LINEAR;
    else
      minFilter = GL_NEAREST;

    // Adjust minFilter if we are using mipmaps
    if (mipmap)
      {
      if (Globals::getValueb("OpenglRenderDevice.minFilterMipmap", TRUE))
        minFilter = (minFilter == GL_LINEAR) ? GL_LINEAR_MIPMAP_LINEAR  : GL_LINEAR_MIPMAP_NEAREST;
      else
        minFilter = (minFilter == GL_LINEAR) ? GL_NEAREST_MIPMAP_LINEAR : GL_NEAREST_MIPMAP_NEAREST;
      }

    // Get magFilter
    if (Globals::getValueb("OpenglRenderDevice.magFilter", TRUE))
      magFilter = GL_LINEAR;
    else
      magFilter = GL_NEAREST;

    // Set texture parameters
    GL_CHECK(glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S,     texture->getWrapS()  ? GL_REPEAT : GL_CLAMP));
    GL_CHECK(glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T,     texture->getWrapT()  ? GL_REPEAT : GL_CLAMP));
    GL_CHECK(glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, minFilter));
    GL_CHECK(glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, magFilter));

    // Set texture mode
    switch (texture->getMode())
      {
      case Texture::REPLACE:
          GL_CHECK(glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE));
          break;

      case Texture::MODULATE:
          GL_CHECK(glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE));
          break;

      case Texture::DECAL:
          GL_CHECK(glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL));
          break;

      case Texture::BLEND:
          GL_CHECK(glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND));
          break;
      }

    // Upload texture
    if (mipmap)
      GL_CHECK(gluBuild2DMipmaps(GL_TEXTURE_2D, 3, image->width, image->height, GL_RGB, GL_UNSIGNED_BYTE, image->data));
    else
      GL_CHECK(glTexImage2D(GL_TEXTURE_2D, 0, 3, image->width, image->height, 0, GL_RGB, GL_UNSIGNED_BYTE, image->data));

    // We dont need this image anymore
    delete image;
    }
  }


RTTI_DEF1(OpenglRenderDevice, "OpenglRenderDevice", RenderDevice);