//
// Copyright (c) 1999, Bernhard Tschirren
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of the source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//
// 2. Any redistribution solely in binary form must conspicuously
//    reproduce the following disclaimer in documentation provided with
//    the binary redistribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'', WITHOUT ANY WARRANTIES, EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  LICENSOR SHALL
// NOT BE LIABLE FOR ANY LOSS OR DAMAGES RESULTING FROM THE USE OF THIS
// SOFTWARE, EITHER ALONE OR IN COMBINATION WITH ANY OTHER SOFTWARE.
//

#include "nodes/PortalNode.h"
#include "render/PortalRenderProcess.h"
#include "core/Frustum.h"
#include "core/Matrix.h"
#include "util/Debug.h"


Boolean PortalNode::preRender(RenderProcess* process_)
  {
  PortalRenderProcess* process = (PortalRenderProcess*)process_;
  RenderDevice*        device  = process->getRenderDevice();  
  PointArray           tempVerts[2];
  PointArray*          outVerts;

  // Check that portal is facing viewer  
  if (_plane.distance(process->getViewer()->getTranslation()) >= 0)
    return FALSE;

  // Transform portal into viewer-space
  // @todo Here I use device->getMatrix() which could be VERY SLOW
  //       on next generation graphics cards. A better approach would
  //       be to maintain our own matrix stack... Maybe :)
  outVerts = transformPortal(_vertices, device->getMatrix(), tempVerts[0]);
  if (!outVerts)
    return FALSE;

  // Clip portal to current view frustum
  outVerts = clipToFrustum(process->getFrustum(), outVerts, tempVerts);
  if (!outVerts)
    return FALSE;

  // Cast new frustum through portal.
  // @todo It would be relatively simple make these frustum use
  //       OpenGL's user-defined clipping planes. This might
  //       give a significant performance improvement (especially
  //       on next generation cards). NOTE: To implement this,
  //       the frustum stack should be moved into the device
  //       (and work very similar to the lights).
  Frustum* frustum = newFrustum(outVerts);
  if (!frustum)
    return FALSE;

  // Set new frustum
  process->pushFrustum(frustum);

  // Lights remain local to current cell
  device->pushLights();
  device->clearLights();

  // This also saves the current model-view matrix
  return RenderNode::preRender(process);
  }


Void PortalNode::postRender(RenderProcess* process_)
  {
  PortalRenderProcess* process = (PortalRenderProcess*)process_;
  RenderDevice*        device  = process->getRenderDevice();

  // Restore the current model-view matrix
  RenderNode::postRender(process);

  // Restore previous frustum
  delete process->popFrustum();

  // Restore lights for current cell
  device->popLights();
  }


PortalNode::PointArray* PortalNode::transformPortal(const PointArray& inVerts, const Matrix& matrix, PointArray& outVerts)
  {
  Int numVerts = inVerts.getLength();

  outVerts.setLength(numVerts);
  for (Int i = 0; i < numVerts; i++)
    outVerts[i] = matrix.multiply(inVerts[i]);

  return &outVerts;
  }


PortalNode::PointArray* PortalNode::clipToPlane(const Plane& plane, const PointArray* inVerts, PointArray* outVerts)
  {
  const Int     numVertices = inVerts->getLength();
  const Point*  curVertex;
        Float   curDot;
        Boolean curIn;

  // Clear the list of clipped vertices
  outVerts->clear();

  // Start with the first vertex
  curVertex = inVerts->getPtr(0);
  curDot    = curVertex->dotProduct(plane.getNormal());
  curIn     = (curDot >= plane.getDistance());

  // Clip all vertices
  for (Int i = 0; i < numVertices; i++)
    {
    // Get next vertex (with wrap-around)
    const Point*  nextVertex = inVerts->getPtr((i + 1) % numVertices);
    const Float   nextDot    = nextVertex->dotProduct(plane.getNormal());
    const Boolean nextIn     = (nextDot >= plane.getDistance());

    // Keep the current vertex if its inside the plane
    if (curIn)
      outVerts->append(*curVertex);

    // Add a clipped vertex if one end of the current edge
    // is inside the plane and the other is outside
    if (curIn != nextIn)
      {
      // Calculate 'time' of intersection
      Float time = (plane.getDistance() - curDot) / (nextDot - curDot);

      // Create a new vertex
      Point newVertex(*curVertex);
      newVertex.interpolate(*nextVertex, time);

      // Add new vertex to list
      outVerts->append(newVertex);
      }

    curDot    = nextDot;
    curIn     = nextIn;
    curVertex = nextVertex;
    }

  return (outVerts->getLength() >= 3) ? outVerts : NULL;
  }


PortalNode::PointArray* PortalNode::clipToFrustum(const Frustum* frustum, const PointArray* inVerts, PointArray outVerts[2])
  {
        Int            outIndex = 1;
  const TArray<Plane>& planes   = frustum->getPlanes();
  const PointArray*    in       = inVerts;
        PointArray*    out      = NULL;

  // Clip to each plane in turn
  for (Int i = 0; i < planes.getLength(); i++)
    {
    out = &outVerts[outIndex];

    if (!clipToPlane(planes[i], in, out))
      {
      out = NULL;
      break;
      }

    in        = out;
    outIndex ^= 1;
    }

  return out;
  }


Frustum* PortalNode::newFrustum(const PointArray* inVerts)
  {
  Frustum* frustum  = new Frustum;
  Int      numVerts = inVerts->getLength();

  for (Int i = 0; i < numVerts; i++)
    frustum->addPlane(Plane(Point(),
                            (*inVerts)[i],
                            (*inVerts)[(i+1) % numVerts]));

  return frustum;
  }


RTTI_DEF1(PortalNode, "PortalNode", RenderNode)