//
// Copyright (c) 1999, Bernhard Tschirren
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of the source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//
// 2. Any redistribution solely in binary form must conspicuously
//    reproduce the following disclaimer in documentation provided with
//    the binary redistribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'', WITHOUT ANY WARRANTIES, EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  LICENSOR SHALL
// NOT BE LIABLE FOR ANY LOSS OR DAMAGES RESULTING FROM THE USE OF THIS
// SOFTWARE, EITHER ALONE OR IN COMBINATION WITH ANY OTHER SOFTWARE.
//

#ifndef _BT_PORTALNODE_H_
#define _BT_PORTALNODE_H_

#include "core/RenderNode.h"
#include "core/Point.h"
#include "core/Plane.h"
#include "core/Frustum.h"
#include "util/TArray.h"

class PortalNode : public RenderNode
  {
  private: typedef TArray<Point> PointArray;

  private: Plane      _plane;
  private: PointArray _vertices;

  public: PortalNode()
            : _plane(),
              _vertices()
            {
            }

  public: PortalNode(const Point* v0, const Point* v1, const Point* v2, ...)
            : _plane(),
              _vertices()
            {
            va_list      args;
            const Point* vertex;

            addVertex(*v0);
            addVertex(*v1);
            addVertex(*v2);

            va_start(args, v2);
            while ((vertex = va_arg(args, const Point*)) != NULL)
              addVertex(*vertex);
            va_end(args);
            }

  public: Void addVertex(const Point& vertex)
            {
            if (_vertices.getLength() == 2)
              {
              _plane.set(_vertices[0],
                         _vertices[1],
                         vertex);
              }
            else
              {
              Float distance = _plane.distance(vertex);

              if (!FLOAT_IS_ZERO(distance))
                throw InvalidArgumentException("Invalid vertex [%f,%f,%f]: "
                                               "It does not lie on plane [%f,%f,%f,%f], "
                                               "distance = %f",
                                               vertex.getX(), vertex.getY(), vertex.getZ(),
                                               _plane.getNormal().getI(), _plane.getNormal().getJ(), _plane.getNormal().getK(), _plane.getDistance(),
                                               distance);
              }

            _vertices.add(vertex);
            }

  public: const Plane& getPlane() const
            {
            return _plane;
            }

  public: const TArray<Point>& getVertices() const
            {
            return _vertices;
            }

  public: virtual Boolean preRender(RenderProcess* process);
  public: virtual Void    postRender(RenderProcess* process);

  private: PointArray* transformPortal(const PointArray& inVerts, const Matrix& matrix, PointArray& outVerts);
  private: PointArray* clipToPlane    (const Plane&   plane,   const PointArray* inVerts, PointArray* outVerts);
  private: PointArray* clipToFrustum  (const Frustum* frustum, const PointArray* inVerts, PointArray outVerts[2]);
  private: Frustum*    newFrustum     (const PointArray* inVerts);

  TYPE_DATA // RTTI
  };

#endif /* _BT_PORTALNODE_H_ */