//
// Copyright (c) 1999, Bernhard Tschirren
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of the source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//
// 2. Any redistribution solely in binary form must conspicuously
//    reproduce the following disclaimer in documentation provided with
//    the binary redistribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'', WITHOUT ANY WARRANTIES, EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  LICENSOR SHALL
// NOT BE LIABLE FOR ANY LOSS OR DAMAGES RESULTING FROM THE USE OF THIS
// SOFTWARE, EITHER ALONE OR IN COMBINATION WITH ANY OTHER SOFTWARE.
//

#include "parsers/Parser.h"


Void Parser::parse()  
  {
  // @todo This is kind of a hack at the moment. Eventually, the parsers
  //       and nodes will be moved off into their own DLL's. Then, each
  //       such DLL will contain a RegisterClasses() function. Until
  //       then, this is all I can do :)
  extern Void RegisterClasses();
  RegisterClasses();

  while (curToken() != tokEOF)
    parseObject(NULL, STATIC_TYPE_INFO(BaseClass));
  }


BaseClass* Parser::find(const String& identifier, RTTItypeid type)
  {
  Int        index;
  BaseClass* object;

  index = findSymbol(identifier);
  if (index < 0)
    throw Exception("Could not find %s!", (const Char*)identifier);

  object = getSymbols()[index].object;
  if (!DYN_CAST(type, object))
    throw Exception("sScene1 is not a Scene!");

  return object;
  }


Int Parser::parseInt()
  {
  Int value;
  Int result;

  expect(tokINTEGER);
  result = sscanf(curString().getPtr(), "%d", &value);
  if (result != 1)
    throw ParserException(this, String().setFormatted("Cannot convert '%s' to an integer (possibly out of range)", curString().getPtr()));
  consume();

  return value;
  }


Float Parser::parseFloat()
  {
  Float value;
  Int   result;

  expect(tokINTEGER, tokFLOAT);
  result = sscanf(curString().getPtr(), "%f", &value);
  if (result != 1)
    throw ParserException(this, String().setFormatted("Cannot convert '%s' to a float (possibly out of range)", curString().getPtr()));
  consume();

  return value;
  }


String Parser::parseString()
  {
  String value;

  expect(tokSTRING);
  value = curString();
  consume();

  return value;
  }


Boolean Parser::parseBoolean()
  {
  Boolean value;

  expect(tokIDENTIFIER);

  if ((curString() == "yes") || (curString() == "on") || (curString() == "true"))
    {
    value = TRUE;
    consume();
    }
  else if ((curString() == "no") || (curString() == "off") || (curString() == "false"))
    {
    value = FALSE;
    consume();
    }
  else
    consume("yes/no");

  return value;
  }


BaseClass* Parser::parseObject(BaseClass* owner, RTTItypeid type)
  {
  String       objectClass;
  String       parserClass;
  ParseObject* parser;
  BaseClass*   object;
  Int          symbol;

  // The identifier specifies the object class
  expect(tokIDENTIFIER);

  // Create Parser (based on class name)
  objectClass  = curString();
  parserClass  = "Parse" + objectClass;
  parser       = (ParseObject*)(STATIC_TYPE_INFO(ParseObject).create(RTTIdyntypeid(parserClass)));
  if (!parser)
    throw ParserException(this, String().setFormatted("Unknown class '%s'", (const Char*)curString()));

  // Consume class name
  consume();

  // Parse identifier
  if (curToken() == tokIDENTIFIER)
    {
    symbol = addSymbol(curString());
    consume(tokIDENTIFIER);
    }
  else
    symbol = -1;
  
  // Parse object
  consume(tokLEFTBRACE);
  object = parser->parse(*this, owner);
  consume(tokRIGHTBRACE);

  if (object != NULL)
    {
    // Check type of object
    if (!DYN_CAST(type, object))
      throw ParserException(this, String().setFormatted("'%s' is not a sub-class of '%s'", (Char*)objectClass, type.getname()));  

    // Now, point symbol at the object
    if (symbol >= 0)
      getSymbols()[symbol].object = object;
    }
  
  // Delete Parser
  delete parser;

  return object;
  }


BaseClass* Parser::parseParent(BaseClass* owner, const String& className)
  {
  String       parserClass;
  ParseObject* parser;
  BaseClass*   object;

  // Create Parser (based on class name)
  parserClass = "Parse" + className;
  parser      = (ParseObject*)(STATIC_TYPE_INFO(ParseObject).create(RTTIdyntypeid(parserClass)));
  if (!parser)
    throw ParserException(this, String().setFormatted("Unknown class '%s'", (const Char*)curString()));

  // Parse object
  object = parser->parse(*this, owner);

  return object;
  }


Void Parser::parseIntArray(TArray<Int>& array)
  {
  consume(tokLEFTBRACKET);

  while (curToken() != tokRIGHTBRACKET)
    array.add(parseInt());

  consume(tokRIGHTBRACKET);
  }


Void Parser::parseFloatArray(TArray<Float>& array)
  { 
  consume(tokLEFTBRACKET);

  while (curToken() != tokRIGHTBRACKET)
    array.add(parseFloat());

  consume(tokRIGHTBRACKET);
  }


Void Parser::parseObjectArray(TArray<BaseClass*>& array, BaseClass* owner, RTTItypeid type)
  {
  consume(tokLEFTBRACKET);

  while (curToken() != tokRIGHTBRACKET)
    array.add(parseObject(owner, type));

  consume(tokRIGHTBRACKET);
  }


Int Parser::findSymbol(const String& identifier) const
  {
  Int found = -1;  

  for (Int i = 0; i < _symbols.getLength(); i++)
    {  
    if (_symbols[i].identifier == identifier)
      {
      found = i;
      break;
      }
    }

  return found;
  }


Int Parser::addSymbol(const String& identifier)
  {
  Symbol symbol;

  Int index = findSymbol(identifier);
  if (index >= 0)
    throw ParserException(this, String().setFormatted("\"%s\" already declared in %s:%d!",
                          (Char*)_symbols[index].identifier,
                          (Char*)_symbols[index].file,
                                 _symbols[index].line));

  symbol.identifier = identifier;
  symbol.object     = NULL;
  symbol.file       = curFile();
  symbol.line       = curLine();

  return _symbols.add(symbol);
  }


Int Parser::addSymbol(const Symbol& other)
  {
  Int index = addSymbol(other.identifier);

  _symbols[index].object = other.object;
  _symbols[index].file   = other.file;
  _symbols[index].line   = other.line;

  return index;
  }


RTTI_DEF1(Parser,      "Parser",      BaseClass)
RTTI_DEF1(ParseObject, "ParseObject", BaseClass)