//
// Copyright (c) 1999, Bernhard Tschirren
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of the source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//
// 2. Any redistribution solely in binary form must conspicuously
//    reproduce the following disclaimer in documentation provided with
//    the binary redistribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'', WITHOUT ANY WARRANTIES, EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  LICENSOR SHALL
// NOT BE LIABLE FOR ANY LOSS OR DAMAGES RESULTING FROM THE USE OF THIS
// SOFTWARE, EITHER ALONE OR IN COMBINATION WITH ANY OTHER SOFTWARE.
//

#include "render/PortalRenderProcess.h"
#include "util/Graph.h"
#include "core/RenderDevice.h"
#include "core/RenderNode.h"
#include "nodes/CellNode.h"


class RenderDFS : public GraphDFS
  {
  private: PortalRenderProcess* _process;
  private: CellNode*            _cell;  

  public: Void execute(PortalRenderProcess* process, CellNode* cell)
            {
            _cell    = cell;
            _process = process;

            GraphDFS::execute(cell);
            }

  private: virtual Boolean preProcess(GraphNode* node)
             {
             return ((RenderNode*)node)->preRender(_process);
             }
           
  private: virtual Void postProcess(GraphNode* node)
             {
             ((RenderNode*)node)->postRender(_process);
             }
  };


static RenderDFS renderDFS;


Void PortalRenderProcess::render(RenderDevice* device)
  {
  setRenderDevice(device);

  // Create initial view frustum
  // @todo This does not havew to be done every frame, only
  //       whenever the user resizes the window!
  newFrustum();

  // Write start-of-scene marker
  device->beginScene();
 
  // Clear some stuff
  device->clearLights();
  device->clear(RenderDevice::CLEAR_COLOR_BUFFER |
                RenderDevice::CLEAR_DEPTH_BUFFER);

  if (getViewer() != NULL)
    {
    RenderNode* node = getViewer();
    CellNode*   cell = NULL;

    // Search for base-cell
    while (cell == NULL)
      {
      // Perform inverse transformation
      device->rotate(-node->getRotation());
      device->translate(-node->getTranslation());

      // Follow first incoming link
      node = node->getParent();
      if (node == NULL)
        {
        setViewer(NULL);
        throw RenderException(*this, "Viewer is not attached to a Cell!");
        }

      // Do we have cell?
      cell = PTR_CAST(CellNode, node);
      }

    // Now render starting from found cell
    renderDFS.execute(this, cell);
    }

  // Write end-of-scene marker
  device->endScene();

  // Delete initial frustum
  deleteFrustum();
  }


Void PortalRenderProcess::newFrustum()
  {
  RenderDevice* device = getRenderDevice();

  // Create initial frustum (based on fovY and aspect).
  // NOTE: Viewer is always located at the origin (0,0,0). Therefore,
  //       coordinates must be converted to viewer-space before they
  //       can be tested against view frustum.
  Float    ymax    = _zNear * tan(_fovY * M_PI / 360.0);
  Float    ymin    = -ymax;
  Float    xmin    = ymin * _aspect * device->w2() / device->h2();
  Float    xmax    = ymax * _aspect * device->w2() / device->h2();
  Frustum* frustum = new Frustum();

  frustum->addPlane(Plane(Point(),
                          Point(xmin, ymin, _zNear),
                          Point(xmin, ymax, _zNear)));

  frustum->addPlane(Plane(Point(),
                          Point(xmin, ymax, _zNear),
                          Point(xmax, ymax, _zNear)));

  frustum->addPlane(Plane(Point(),
                          Point(xmax, ymax, _zNear),
                          Point(xmax, ymin, _zNear)));

  frustum->addPlane(Plane(Point(),
                          Point(xmax, ymin, _zNear),
                          Point(xmin, ymin, _zNear)));

  // Set devices's frustum
  device->setPerspective(xmin, xmax, ymin, ymax, _zNear, _zFar);

  // Make sure stack is empty
  ASSERT(getFrustum() == NULL);

  pushFrustum(frustum);
  }


Void PortalRenderProcess::deleteFrustum()
  {
  delete popFrustum();

  // Make sure stack is empty
  ASSERT(getFrustum() == NULL);
  }


RTTI_DEF1(PortalRenderProcess, "PortalRenderProcess", RenderProcess)