//
// Copyright (c) 1999, Bernhard Tschirren
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of the source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//
// 2. Any redistribution solely in binary form must conspicuously
//    reproduce the following disclaimer in documentation provided with
//    the binary redistribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'', WITHOUT ANY WARRANTIES, EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  LICENSOR SHALL
// NOT BE LIABLE FOR ANY LOSS OR DAMAGES RESULTING FROM THE USE OF THIS
// SOFTWARE, EITHER ALONE OR IN COMBINATION WITH ANY OTHER SOFTWARE.
//

#include "util/Globals.h"
#include "util/Scanner.h"


TArray<Globals::Value> Globals::_values;


Boolean Globals::getValueb(const String& name, Boolean defaultValue)
  {
  Boolean found;
  Int     index;
    
  found = find(name, 'b', &index);

  if (found)
    return _values[index].valueb;
  else
    return defaultValue;
  }


Void Globals::setValueb(const String& name, Boolean value)
  {
  Boolean found;
  Int     index;
  
  found = find(name, 'b', &index);

  if (found)
    _values[index].valueb = value;
  else
    {
    Value element;

    element.name   = name;
    element.type   = 'b';
    element.valueb = value;

    _values.insert(index, element);
    }
  }


Int Globals::getValuei(const String& name, Int defaultValue)
  {
  Boolean found;
  Int     index;
    
  found = find(name, 'i', &index);

  if (found)
    return _values[index].valuei;
  else
    return defaultValue;
  }


Void Globals::setValuei(const String& name, Int value)
  {
  Boolean found;
  Int     index;
  
  found = find(name, 'i', &index);

  if (found)
    _values[index].valuei = value;
  else
    {
    Value element;

    element.name   = name;
    element.type   = 'i';
    element.valuei = value;

    _values.insert(index, element);
    }
  }


Float Globals::getValuef(const String& name, Float defaultValue)
  {
  Boolean found;
  Int     index;
    
  found = find(name, 'f', &index);

  if (found)
    return _values[index].valuef;
  else
    return defaultValue;
  }


Void Globals::setValuef(const String& name, Float value)
  {
  Boolean found;
  Int     index;
  
  found = find(name, 'f', &index);

  if (found)
    _values[index].valuef = value;
  else
    {
    Value element;

    element.name   = name;
    element.type   = 'f';
    element.valuef = value;

    _values.insert(index, element);
    }
  }


String Globals::getValues(const String& name, const String& defaultValue)
  {
  Boolean found;
  Int     index;
    
  found = find(name, 's', &index);

  if (found)
    return *(_values[index].values);
  else
    return defaultValue;
  }


Void Globals::setValues(const String& name, const String& value)
  {
  Boolean found;
  Int     index;
  
  found = find(name, 's', &index);

  if (found)
    *(_values[index].values) = value;
  else
    {
    Value element;

    element.name   = name;
    element.type   = 's';
    element.values = new String(value);

    _values.insert(index, element);
    }
  }


Void Globals::load(FileStream& stream)
  {
  Scanner scanner(&stream);
  String  name;

  while (scanner.curToken() != Scanner::tokEOF)
    {
    // Parse identifier
    scanner.expect(Scanner::tokIDENTIFIER);
    name = scanner.curString();
    scanner.consume(Scanner::tokIDENTIFIER);

    // Identifier may have periods in them
    while (scanner.curToken() == Scanner::tokPERIOD)
      {
      scanner.consume(Scanner::tokPERIOD);

      scanner.expect(Scanner::tokIDENTIFIER);
      name += '.';
      name += scanner.curString();
      scanner.consume(Scanner::tokIDENTIFIER);
      }

    // Get equal sign
    scanner.consume(Scanner::tokEQUAL);

    // Get value
    switch (scanner.curToken())
      {
      // Parse Boolean value
      case Scanner::tokIDENTIFIER:
        if (scanner.curString() == "true")
          {
          setValueb(name, TRUE);
          scanner.consume();
          }
        else if (scanner.curString() == "false")
          {
          setValueb(name, FALSE);
          scanner.consume();
          }
        else
          scanner.consume("true or false");        
        break;

      // Parse Int value
      case Scanner::tokINTEGER:
        setValuei(name, atoi(scanner.curString()));
        scanner.consume();
        break;

      // Parse Float value
      case Scanner::tokFLOAT:
        setValuef(name, atof(scanner.curString()));
        scanner.consume();
        break;

      // Parse String value
      case Scanner::tokSTRING:
        setValues(name, scanner.curString());
        scanner.consume();
        break;
      }
    }
  }


Void Globals::save(FileStream& stream)
  {
  for (Int i = 0; i < _values.getLength(); i++)
    {
    switch (_values[i].type)
      {
      // Write Boolean value
      case 'b':
        stream.printf("%s = %s\n",
                      (const Char*)_values[i].name,
                      (_values[i].valueb) ? "true" : "false");
        break;

      // Write Int value
      case 'i':
        stream.printf("%s = %d\n",
                      (const Char*)_values[i].name,
                      _values[i].valuei);
        break;

      // Write Float value
      case 'f':
        stream.printf("%s = %e\n",
                      (const Char*)_values[i].name,
                      _values[i].valuef);
        break;

      // Write String value
      case 's':
        stream.printf("%s = \"%s\"\n",
                      (const Char*)_values[i].name,
                      (const Char*)(*_values[i].values));
        break;
      }
    }
  }


const Char* Globals::typeName(Char type)
  {
  switch (type)
    {
    case 'b': return "Boolean";
    case 'i': return "Int";
    case 'f': return "Float";
    case 's': return "String";
    default:  return "Unknown";
    }
  }


Boolean Globals::find(const String& name, Char type, Int* index)
  {
  Boolean found = FALSE;
  Int     left  = 0;
  Int     right = _values.getLength()-1;  

  while (left <= right)
    {
    Int middle = (left + right) / 2;
    Int comp   = name.compare(_values[middle].name);

    if (comp > 0)
      left  = middle+1;

    else if (comp < 0)
      right = middle-1;

    else
      {
      // Found a match, so check type
      if (_values[middle].type != type)
        throw Exception("Globals type mismatch: %s is '%s', but it should be '%s'!",
                        (const Char*)_values[middle].name,
                        typeName(_values[middle].type),
                        typeName(type));

      // Return index number
      found = TRUE;
      left  = middle;
      break;
      }
    }

  if (index) *index = left;
  return found;
  }