//
// Copyright (c) 1999, Bernhard Tschirren
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of the source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//
// 2. Any redistribution solely in binary form must conspicuously
//    reproduce the following disclaimer in documentation provided with
//    the binary redistribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'', WITHOUT ANY WARRANTIES, EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  LICENSOR SHALL
// NOT BE LIABLE FOR ANY LOSS OR DAMAGES RESULTING FROM THE USE OF THIS
// SOFTWARE, EITHER ALONE OR IN COMBINATION WITH ANY OTHER SOFTWARE.
//

#ifndef _BT_GRAPH_H_
#define _BT_GRAPH_H_

#include "util/BaseClass.h"
#include "util/String.h"
#include "util/TArray.h"
class GraphNode;

typedef TArray<GraphNode*> GraphNodeArray;

class GraphNode : public BaseClass
  {
  private: GraphNodeArray _parents;
  private: GraphNodeArray _children;

  public: GraphNode()
            : _parents(),
              _children()
            {
            }

  public: virtual ~GraphNode()
            {
            }

  public: Int getNumParents() const
            {
            return _parents.getLength();
            }

  public: const GraphNodeArray& getParents() const
            {
            return _parents;
            };

  public: Boolean parentsContains(const GraphNode* node) const
            {
            return (_parents.find((GraphNode*)node) >= 0);
            };

  public: Int getNumChildren() const
            {
            return _children.getLength();
            };

  public: const GraphNodeArray& getChildren() const
            {
            return _children;
            };

  public: Boolean childrenContains(const GraphNode* node) const
            {
            return (_children.find((GraphNode*)node) >= 0);
            };

  public: Void attachParent(GraphNode* parent)
            {
            _parents.add(parent);
            parent->_children.add(this);
            }

  public: Void removeParent(GraphNode* parent)
            {
            _parents.remove(parent);
            parent->_children.remove(this);
            }

  public: Void attachChild(GraphNode* child)
            {
            _children.add(child);
            child->_parents.add(this);
            }

  public: Void detachChild(GraphNode* child)
            {
            _children.remove(child);
            child->_parents.remove(this);
            }

  TYPE_DATA // RTTI
  };


class Graph : public BaseClass
  {
  private: GraphNodeArray _nodes;

  public: Graph()
            : _nodes()
            {
            }

  public: virtual ~Graph()
            {
            for (Int i = 0; i < _nodes.getLength(); i++)
              delete _nodes[i];
            }

  public: Void clear()
            {
            _nodes.clear();
            }

  public: Int getNumNodes() const
            {
            return _nodes.getLength();
            }

  public: const GraphNodeArray& getNodes() const
            {
            return _nodes;
            }

  public: Boolean contains(const GraphNode* node) const
            {
            return (_nodes.find((GraphNode*)node) >= 0);
            }

  public: Int add(GraphNode* node)
            {
            return _nodes.addUnique(node);
            }

  public: Void remove(GraphNode* node)
            {
            _nodes.remove(node);
            }

  TYPE_DATA // RTTI
  };


class GraphDFS : public BaseClass
  {
  private: GraphNodeArray _processed;

  public:  GraphDFS()
             : _processed()
             {
             }

  public:  virtual ~GraphDFS()
             {
             }

  public:  Void execute(GraphNode* startNode);
  private: Void executeInternal(GraphNode* node);

  private: virtual Boolean preProcess(GraphNode* node)  { return TRUE; };
  private: virtual Void    postProcess(GraphNode* node) {              };

  TYPE_DATA // RTTI
  };


#endif /* _BT_GRAPH_H_ */