//
// Copyright (c) 1999, Bernhard Tschirren
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of the source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//
// 2. Any redistribution solely in binary form must conspicuously
//    reproduce the following disclaimer in documentation provided with
//    the binary redistribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'', WITHOUT ANY WARRANTIES, EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  LICENSOR SHALL
// NOT BE LIABLE FOR ANY LOSS OR DAMAGES RESULTING FROM THE USE OF THIS
// SOFTWARE, EITHER ALONE OR IN COMBINATION WITH ANY OTHER SOFTWARE.
//

#include "util/Memory.h"
#include <stdarg.h>


const Int Memory::GUARD               = 0xDEADBEEF;
const Int Memory::LEADING_GUARD_SIZE  = sizeof(Int)*2; // SIZE, LOW_GUARD
const Int Memory::TRAILING_GUARD_SIZE = sizeof(Int)*1; // HIGH_GUARD
const Int Memory::TOTAL_GUARD_SIZE    = Memory::LEADING_GUARD_SIZE + Memory::TRAILING_GUARD_SIZE;
      Int Memory::_memoryUsed         = 0;
      Int Memory::_maxMemoryUsed      = 0;


Int Memory::getPageSize()
  {
#if defined(_WIN32)
  SYSTEM_INFO si;
  GetSystemInfo(&si);
  return si.dwPageSize;

#elif defined(_UNIX)
  return getpagesize();

#else
  return 4096;

#endif
  }


Int Memory::getVirtualAvailable()
  {
  return 0;
  }


Int Memory::getPhysicalAvailable()
  {
  return 0;
  }


Void* Memory::allocate(Int size, const Char* srcName, Int srcLine)
  {
  ASSERT_SRC(size >= 0, srcName, srcLine);

  Byte* bytePointer;  

  bytePointer = (Byte*)::malloc(size + TOTAL_GUARD_SIZE);
  if (!bytePointer)
    throw OutOfMemoryException("%s%d: Memory::allocate(): Could not allocate %d bytes of memory", srcName, srcLine, size);

  bytePointer   += LEADING_GUARD_SIZE;
  _memoryUsed   += size;
  _maxMemoryUsed = MAX(_memoryUsed, _maxMemoryUsed);

  *getSizePtr(bytePointer)      = size;
  *getLowGuardPtr(bytePointer)  = GUARD;
  *getHighGuardPtr(bytePointer) = GUARD;

  return bytePointer;
  }


Void* Memory::reallocate(Void* pointer, Int size, const Char* srcName, Int srcLine)
  {
  ASSERT_SRC(size >= 0, srcName, srcLine);

  Byte* bytePointer;
  Int   oldSize;

  if (pointer)
    {
    bytePointer = (Byte*)pointer;
    oldSize     = getMemorySize(pointer);

    checkGuards(pointer, srcName, srcLine);

    bytePointer = (Byte*)::realloc(bytePointer - LEADING_GUARD_SIZE, size + TOTAL_GUARD_SIZE);
    if (!bytePointer)
      throw OutOfMemoryException("%s%d: Memory::reallocate(): Could not reallocate %d bytes of memory", srcName, srcLine, size);

    bytePointer   += LEADING_GUARD_SIZE;
    _memoryUsed   -= oldSize;
    _memoryUsed   += size;
    _maxMemoryUsed = MAX(_memoryUsed, _maxMemoryUsed);

    *getSizePtr(bytePointer)      = size;
    *getLowGuardPtr(bytePointer)  = GUARD;
    *getHighGuardPtr(bytePointer) = GUARD;
    }
  else
    bytePointer = (Byte*)allocate(size, srcName, srcLine);

  return bytePointer;
  }


Void Memory::free(Void* pointer, const Char* srcName, Int srcLine)
  { 
  if (pointer)
    {
    checkGuards(pointer, srcName, srcLine);

    _memoryUsed -= getMemorySize(pointer);
    ::free((Byte*)pointer - LEADING_GUARD_SIZE);
    }
  }


OutOfMemoryException::OutOfMemoryException(const Char* format, ...)
  {
  va_list args;

  va_start(args, format);
  vsprintf(_message, format, args);
  va_end(args);
  }