//
// Copyright (c) 1999, Bernhard Tschirren
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of the source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//
// 2. Any redistribution solely in binary form must conspicuously
//    reproduce the following disclaimer in documentation provided with
//    the binary redistribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'', WITHOUT ANY WARRANTIES, EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  LICENSOR SHALL
// NOT BE LIABLE FOR ANY LOSS OR DAMAGES RESULTING FROM THE USE OF THIS
// SOFTWARE, EITHER ALONE OR IN COMBINATION WITH ANY OTHER SOFTWARE.
//

#include "util/Scanner.h"

#define END_OF_LINE         '\n'
#define END_OF_FILE         '\032'

#define IS_END_OF_LINE(c)   ((c) == END_OF_LINE)
#define IS_END_OF_FILE(c)   ((c) == END_OF_FILE)
#define IS_IGNORED(c)       (!IS_END_OF_FILE(c) && (((c) <= 32) || ((c) >= 127)))
#define IS_LETTER(c)        ((((c) >= 'A') && ((c) <= 'Z')) ||  (((c) >= 'a') && ((c) <= 'z')))
#define IS_NUMBER(c)        (((c) >= '0') && ((c) <= '9'))
#define IS_IDENTIFIER(c)    (IS_LETTER(c) || IS_NUMBER(c) || ((c) == '_') || ((c) == '*'))
#define IS_PLUS_SIGN(c)     ((c) == '+')
#define IS_MINUS_SIGN(c)    ((c) == '-')
#define IS_DECIMAL_POINT(c) ((c) == '.')
#define IS_EXPONENT(c)      (((c) == 'e') || ((c) == 'E'))

const Char* ScannerTokenNames[] =
  {
  "invalid character",
  "end of file",
  "comment",
  "identifier",
  "string",
  "floating point number",
  "integer",
  "equal (=)",
  "period (.)",
  "comma (,)",
  "colon (:)",
  "semicolon (;)",
  "left brace ({)",
  "right brace (})",
  "left brackets ([)",
  "right bracket (])"
  };


Scanner::Scanner(FileStream* stream)
  : _stream(*stream),
    _line(1),
    _string(),
    _token(tokINVALID)
  {
  scanChar();
  scanToken();
  }


Scanner::Token Scanner::scanIdentifier()
  {
  do
    {
    _string += _char;
    scanChar();
    } while (IS_IDENTIFIER(_char));

  return tokIDENTIFIER;
  }


Scanner::Token Scanner::scanNumber()
  {
  Scanner::Token result = tokINVALID;

  // Grab integer part (numbers before decimal point)
  if (IS_PLUS_SIGN(_char)  ||
      IS_MINUS_SIGN(_char) ||
      IS_NUMBER(_char))
    {
    result = tokINTEGER;
    do
      {
      _string += _char;
      scanChar();
      } while (IS_NUMBER(_char));
    }

  // Grab fractional part (numbers after decimal point)
  if (IS_DECIMAL_POINT(_char))
    {
    result = tokFLOAT;
    do
      {
      _string += _char;
      scanChar();
      } while (IS_NUMBER(_char));
    }

  // Grab exponent (numbers after 'e')
  if (IS_EXPONENT(_char))
    {
    result = tokFLOAT;

    // Add 'e' to string
    _string += _char;
    scanChar();

    // Add optional '+', '-' to string
    if (IS_PLUS_SIGN(_char) || IS_MINUS_SIGN(_char))
      {
      _string += _char;
      scanChar();
      }

    // Make sure the mandatory exponent follows
    if (!IS_NUMBER(_char))
      throw ScannerException(this, "Expecting exponent");

    // Add exponent
    do
      {
      _string += _char;
      scanChar();
      } while (IS_NUMBER(_char));
    }

  return result;
  }


Scanner::Token Scanner::scanString()
  {
  do
    {
    do
      {
      scanChar();
      if (IS_END_OF_LINE(_char) || IS_END_OF_FILE(_char))
        throw ScannerException(this, "String exceeds end of line");

      _string += _char;
      } while (_char != '\"');

    scanChar();
    } while (_char == '\"');

  _string.setLength(_string.getLength()-1);
    
  return tokSTRING;
  }


Scanner::Token Scanner::scanComment()
  {
  do
    {
    scanChar();
    } while (!IS_END_OF_LINE(_char) && !IS_END_OF_FILE(_char));
  scanChar();

  return tokCOMMENT;
  }


Char Scanner::scanChar()
  {
  if (_stream.endOfStream())
    _char = END_OF_FILE;
  else
    _stream.readChar(&_char);

  if (IS_END_OF_LINE(_char))
    _line++;

  return _char;
  }


Scanner::Token Scanner::scanToken()
  {
  do
    {
    _string = "";

    while (IS_IGNORED(_char))
      scanChar();

    switch (_char)
      {
      case 'A': case 'B': case 'C': case 'D': case 'E': case 'F': case 'G': case 'H': case 'I':
      case 'J': case 'K': case 'L': case 'M': case 'N': case 'O': case 'P': case 'Q': case 'R':
      case 'S': case 'T': case 'U': case 'V': case 'W': case 'X': case 'Y': case 'Z':
      case 'a': case 'b': case 'c': case 'd': case 'e': case 'f': case 'g': case 'h': case 'i':
      case 'j': case 'k': case 'l': case 'm': case 'n': case 'o': case 'p': case 'q': case 'r':
      case 's': case 't': case 'u': case 'v': case 'w': case 'x': case 'y': case 'z':
      case '_': case '*':
        _token = scanIdentifier();
        break;

      case '0': case '1': case '2': case '3': case '4':
      case '5': case '6': case '7': case '8': case '9':
      case '+': case '-':
        _token = scanNumber();
        break;

      case '\"':
        _token = scanString();
        break;

      case '#':
        _token = scanComment();
        break;

      case '=':
        scanChar();
        _token = tokEQUAL;
        break;

      case '.':
        scanChar();
        _token = tokPERIOD;
        break;

      case ',':
        scanChar();
        _token = tokCOMMA;
        break;

      case ':':
        scanChar();
        _token = tokCOLON;
        break;

      case ';':
        scanChar();
        _token = tokSEMICOLON;
        break;

      case '{':
        scanChar();
        _token = tokLEFTBRACE;
        break;

      case '}':
        scanChar();
        _token = tokRIGHTBRACE;
        break;

      case '[':
        scanChar();
        _token = tokLEFTBRACKET;
        break;

      case ']':
        scanChar();
        _token = tokRIGHTBRACKET;
        break;

      case END_OF_FILE:
        _token = tokEOF;
        break;

      default:
        _token = tokINVALID;        
      }
    } while (_token == tokCOMMENT);

  if (_token == tokINVALID)
    throw ScannerException(this, String().setFormatted("Illegal character (#%d)", (Int)_char));

  return _token;
  }


Void Scanner::expect(Scanner::Token token)
  {
  if (_token != token)
    throw ScannerException(this, String().setFormatted("Expecting %s", tokenName(token)));
  }


Void Scanner::expect(Scanner::Token token1, Scanner::Token token2)
  {
  if ((_token != token1) && (_token != token2))
    throw ScannerException(this, String().setFormatted("Expecting %s or %s", tokenName(token1), tokenName(token2)));
  }


Void Scanner::consume()
  {
  scanToken();
  }


Void Scanner::consume(Scanner::Token token)
  {
  expect(token);
  scanToken();
  }


Void Scanner::consume(const Char* identifier)
  {
  expect(tokIDENTIFIER);
  if (_string != identifier)
    throw ScannerException(this, String().setFormatted("Expecting identifier \"%s\"", identifier));
  scanToken();
  }

const Char* Scanner::tokenName(Scanner::Token token) const
  {
  return ScannerTokenNames[token];
  }


RTTI_DEF1(Scanner, "Scanner", BaseClass)