//
// Copyright (c) 1999, Bernhard Tschirren
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of the source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//
// 2. Any redistribution solely in binary form must conspicuously
//    reproduce the following disclaimer in documentation provided with
//    the binary redistribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'', WITHOUT ANY WARRANTIES, EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  LICENSOR SHALL
// NOT BE LIABLE FOR ANY LOSS OR DAMAGES RESULTING FROM THE USE OF THIS
// SOFTWARE, EITHER ALONE OR IN COMBINATION WITH ANY OTHER SOFTWARE.
//

#ifndef _BT_STRING_H_
#define _BT_STRING_H_

#include "util/TArray.h"
#include <memory.h>
#include <string.h>
#include <ctype.h>

typedef TArray<Char> CharArray;

class String : public CharArray
  {
  private: enum { INITIAL_LENGTH = 64 };

  public: String()
            : CharArray(INITIAL_LENGTH)
            {
            }

  public: String(const String& string)
            : CharArray(string)
            {
            }

  public: String(const Char* string)
            : CharArray(strlen(string)+1, INITIAL_LENGTH)
            {
            set(string);
            }

  public: String(const Char string)
            : CharArray(2, INITIAL_LENGTH)
            {
            set(string);
            }

  public: String(const Char ch, const Int count)
            : CharArray(count+1, 64)
            {
            set(ch, count);
            }

  public: Int getLength() const
            {
            return CharArray::getLength()-1;
            };

  public: String& setLength(Int length)
            {
            CharArray::setLength(length+1);
            *getPtr(length) = '\0';

            return *this;
            }

  public: String& set(const Char* string)
            {
            Int length = strlen(string)+1;

            CharArray::setLength(length);
            memmove(getPtr(), string, length*sizeof(Char));

            return *this;
            }

  public: String& set(const Char string)
            {
            CharArray::setLength(2);
            *getPtr(0) = string;
            *getPtr(1) = '\0';

            return *this;
            }

  public: String& set(const Char ch, Int count)
            {
            Char* ptr;

            CharArray::setLength(count+1);

            ptr = getPtr();
            while (count--) *ptr++ = ch;
            *ptr = '\0';

            return *this;
            }

  public: String& setFormatted(const Char* format, ...)
            {
            va_list args;

            va_start(args, format);
            setFormattedV(format, args);
            va_end(args);

            return *this;
            }

  public: String& setFormattedV(const Char* format, const va_list args)
            {
            // @todo VERY DODGY
            Char buffer[65535];
            vsprintf(buffer, format, args);

            set(buffer);

            return *this;
            }

  public: String& toUpper()
            {
            for (Char* c = getPtr(); *c != '\0'; c++)
              *c = toupper(*c);

            return *this;
            }

  public: String& toLower()
            {
            for (Char* c = getPtr(); *c != '\0'; c++)
              *c = tolower(*c);

            return *this;
            }

  public: Int compare(const Char* string) const
            {
            return strcmp(getPtr(), string);
            }

  public: Int compare(const Char string) const
            {
            return ((getPtr()[0] == string) &&
                    (getPtr()[1] == '\0'));
            }

  public: String& append(const Char* string)
            {
            Int thisLength   = getLength();
            Int stringLength = strlen(string);

            CharArray::setLength(thisLength + stringLength + 1);
            memmove(getPtr()+thisLength, string, (stringLength+1)*sizeof(Char));
  
            return *this;
            }

  public: String& append(const Char string)
            {
            Int thisLength = getLength();
  
            CharArray::setLength(thisLength + 2);
            *getPtr(thisLength)   = string;
            *getPtr(thisLength+1) = '\0';

            return *this;
            }

  public: Boolean contains(const Char* string) const
            {
            return (strstr(getPtr(), string) != NULL);
            }

  public: Boolean contains(const Char string) const
            {
            return (strchr(getPtr(), string) != NULL);
            }

  public: Boolean operator ==(const String& string) const { return (compare(string) == 0); }
  public: Boolean operator ==(const Char* string) const   { return (compare(string) == 0); }
  public: Boolean operator ==(const Char string) const    { return (compare(string) == 0); }

  public: Boolean operator !=(const String& string) const { return (compare(string) != 0); }
  public: Boolean operator !=(const Char* string) const   { return (compare(string) != 0); }
  public: Boolean operator !=(const Char string) const    { return (compare(string) != 0); }

  public: Boolean operator <=(const String& string) const { return (compare(string) <= 0); }
  public: Boolean operator <=(const Char* string) const   { return (compare(string) <= 0); }
  public: Boolean operator <=(const Char string) const    { return (compare(string) <= 0); }

  public: Boolean operator >=(const String& string) const { return (compare(string) >= 0); }
  public: Boolean operator >=(const Char* string) const   { return (compare(string) >= 0); }
  public: Boolean operator >=(const Char string) const    { return (compare(string) >= 0); }

  public: Boolean operator >(const String& string) const  { return (compare(string) >  0); }
  public: Boolean operator >(const Char* string) const    { return (compare(string) >  0); }
  public: Boolean operator >(const Char string) const     { return (compare(string) >  0); }

  public: Boolean operator <(const String& string) const  { return (compare(string) <  0); }
  public: Boolean operator <(const Char* string) const    { return (compare(string) <  0); }
  public: Boolean operator <(const Char string) const     { return (compare(string) <  0); }

  public: const String& operator =(const String& string)  { return set(string); }
  public: const String& operator =(const Char* string)    { return set(string); }
  public: const String& operator =(const Char string)     { return set(string); }

  public: String& operator +=(const String& string)       { return append(string); }
  public: String& operator +=(const Char* string)         { return append(string); }
  public: String& operator +=(Char string)                { return append(string); }

  public: friend String operator +(const Char*   s1, const String& s2) { return String(s1).append(s2); }
  public: friend String operator +(const Char    s1, const String& s2) { return String(s1).append(s2); }
  public: friend String operator +(const String& s1, const String& s2) { return String(s1).append(s2); }
  public: friend String operator +(const String& s1, const Char*   s2) { return String(s1).append(s2); }
  public: friend String operator +(const String& s1, const Char    s2) { return String(s1).append(s2); }
  };

#endif /* _BT_STRING_H_ */