//
// Copyright (c) 1999, Bernhard Tschirren
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of the source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//
// 2. Any redistribution solely in binary form must conspicuously
//    reproduce the following disclaimer in documentation provided with
//    the binary redistribution.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'', WITHOUT ANY WARRANTIES, EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  LICENSOR SHALL
// NOT BE LIABLE FOR ANY LOSS OR DAMAGES RESULTING FROM THE USE OF THIS
// SOFTWARE, EITHER ALONE OR IN COMBINATION WITH ANY OTHER SOFTWARE.
//

#ifndef _BT_TARRAY_H_
#define _BT_TARRAY_H_

#include "util/BaseClass.h"
#include "util/Memory.h"

template<typename T> class TArray : public BaseClass
  {
  private: typedef T        Type;
  private: typedef T*       Pointer;
  private: typedef const T* ConstPointer;
  private: typedef T&       Reference;
  private: typedef const T& ConstReference;

  private: Int     _modCount;
  private: Int     _length;
  private: Int     _capacity;
  private: Int     _granularity;
  private: Pointer _elements;

  public: TArray()
            : _modCount(0),
              _length(0),
              _capacity(0),
              _granularity(-1),
              _elements(NULL)
            {
            }

  public: TArray(const TArray<T>& other)
            : _modCount(0),
              _length(0),
              _capacity(0),
              _granularity(-1),
              _elements(NULL)
            {
            set(other);
            }

  public: TArray(Int capacity, Int granularity=-1)
            : _modCount(0),
              _length(0),
              _capacity(0),
              _granularity(granularity),
              _elements(NULL)
            {
            setCapacity(capacity, (granularity != 0));
            }

  public: ~TArray()
            {
            if (_elements != NULL)
              delete[] _elements;
            }

  public: Boolean isEmpty() const
            {
            return (getLength() <= 0);
            }

  public: Int getLength() const
            {
            return _length;
            }

  public: Int getCapacity() const
            {
            return _capacity;
            }

  public: Int getGranularity() const
            {
            return _granularity;
            }

  public: Int getModCount() const
            {
            return _modCount;
            };

  public: ConstPointer getPtr(Int index=0) const
            {
            ASSERT(index >= 0);
            ASSERT(index < _length);
            return (_elements + index);
            }

  public: Pointer getPtr(Int index=0)
            {
            ASSERT(index >= 0);
            ASSERT(index < _length);
            return (_elements + index);
            }

  public: Int find(ConstReference object, Int startIndex=0) const
            {
            for (Int index = startIndex; index < _length; index++)
              {
              if (_elements[index] == object)
                return index;
              }

            return -1;
            }

  public: Boolean contains(ConstReference object, Int startIndex=0) const
            {
            return (find(object, startIndex) >= 0);
            }

  public: Void clear()
            {
            setLength(0);
            }

  public: Void setLength(Int length)
            {
            ASSERT(length >= 0);
            if (_capacity < length)
              setCapacity(length);
            _length = length;
            }

  public: Void setCapacity(Int capacity, Boolean align=TRUE)
            {
            if (_granularity == -1)
              {
              _granularity = Memory::getPageSize() / sizeof(T);
              if (_granularity < 2) _granularity = 2;
              }

            if (align)
              {
              ASSERT(_granularity > 0);
              capacity = ((capacity + _granularity - 1) / _granularity) * _granularity;
              }            

            if (_capacity != capacity)
              {
              _capacity = capacity;
              ASSERT(_length <= _capacity);

              if (_elements == NULL)
                _elements = new T[_capacity];
              else
                {
                Pointer prev_elements = _elements;

                _elements = new T[_capacity];
                for (Int i = 0; i < _length; i++)
                  _elements[i] = prev_elements[i];

                delete[] prev_elements;
                }

              _modCount++;
              }
            }

  public: Void setGranularity(Int granularity)
            {
            ASSERT(granularity >= -1);

            _granularity = granularity;
            if (granularity != 0)
              setCapacity(capacity);
            }

  public: Void set(Int index, ConstReference data)
            {
            ASSERT(index >= 0);
            ASSERT(index < _length);

            _elements[index] = data;
            _modCount++;
            }

  public: Void set(const TArray<T>& other)
            {
            _granularity = other.getGranularity();

            setCapacity(other.getCapacity(), FALSE);
            setLength(other.getLength());

            for (Int i = 0; i < _length; i++)
              _elements[i] = other[i];

            _modCount++;
            }

  public: Void remove(Int index)
            {
            Int length_1 = getLength() - 1;

            // Shift array elements over the top of removed object
            if (index < length_1)
              {
              for (Int i = index; i < length_1; i++)
                _elements[i] = _elements[i+1];
              }

            // Reduce the length of the array
            setLength(length_1);

            // Let any iterators know that we changed!
            _modCount++;
            }

  public: Void remove(ConstReference data)
            {
            Int index;

            index = find(data);
            while (index >= 0)
              {
              remove(index);
              index = find(data, index);
              }
            }

  public: Int insert(Int index, ConstReference data)
            {
            Int length = getLength();

            // Increase the length of the array
            setLength(length+1);

            // Shift array elements to make room for new element
            if (index < length)
              {
              for (Int i = length; i > index; i--)
                _elements[i] = _elements[i-1];
              }

            // Set new element
            set(index, data);

            // Return index to where we inserted the new element
            return index;
            }

  public: Int prepend(ConstReference data)
            {
            return insert(0, data);
            }

  public: Int append(ConstReference data)
            {
            return insert(_length, data);
            }

  public: Int add(ConstReference data)
            {
            return append(data);
            }

  public: Int addUnique(ConstReference data)
            {
            if (contains(data))
              return -1;
            else
              return append(data);
            }

  public: Void compact()
            {
            setCapacity(_length, FALSE);
            }

  public: ConstReference operator[](Int index) const
            {
            ASSERT(index >= 0);
            ASSERT(index < _length);

            return _elements[index];
            }

  public: Reference operator [](Int index)
            {
            ASSERT(index >= 0);
            ASSERT(index < _length);

            _modCount++;

            return _elements[index];
            }

  public: operator ConstPointer() const
            {
            return getPtr();
            }

  public: operator Pointer()
            {
            return getPtr();
            }

  public: const TArray<T>& operator =(const TArray<T>& other)
            {
            set(other);
            return *this;
            }
  };

#endif /* _BT_TARRAY_H_ */